/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * Raymer, D. P. - Aircraft Design: A Conceptual Approach, American Institute of Aeronautics and Astronautics, 2018, ISBN: 978-1-62410-574-6
 * p. 280 ff. (pdf p. 294 ff.) 
 */

#include "viscDragForTandem.h"

#include <aixml/node.h>
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <moduleBasics/module.h>
#include <moduleBasics/runtimeIO.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/strategySelector.h>

#include <filesystem>
#include <vector>
#include <memory>
#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <utility>  // make_pair

#include "../polar.h"
#include "generalMethods/auxFunctions.h"

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::isinf;
using std::isnan;
using std::vector;
/* TODO:    -find way to substract fuselage from wetted surface wing
            -find way to calculate fuselage upsweep angle
            -check if wing drag is for entire wing or just right side
*/
viscDragForTandemInitialization::viscDragForTandemInitialization()
    :
    calculateWing(false),
    calculateFuselage(false),
    calculateVTP(false),
    calculateBackWing(false),
    calculateNacelles(false),
    calculateMisc(false),
    manualTransitionLocation(false),
    transitionLocationWing(NAN),
    transitionLocationBackWing(NAN),
    transitionLocationVTP(NAN),
    doCalibration(false),
    CDvisc_CDsum(NAN),
    CDvisc_CLfact(NAN),
    CDvisc_CLexp(NAN),
    CDvisc_CDsum_lowMa(NAN),
    CDvisc_CLfact_lowMa(NAN),
    CDvisc_CLexp_lowMa(NAN),
    jetEngine(false) {
}

viscDragForTandem::viscDragForTandem(const std::shared_ptr<RuntimeIO>& rtIO, const viscDragForTandemInitialization& viscDragInit)
    :
    // the following variables need to be initialized externally
    calculateWing(viscDragInit.calculateWing),
    calculateFuselage(viscDragInit.calculateFuselage),
    calculateVTP(viscDragInit.calculateVTP),
    calculateBackWing(viscDragInit.calculateBackWing),
    calculateNacelles(viscDragInit.calculateNacelles),
    calculateMisc(viscDragInit.calculateMisc),
    manualTransitionLocation(viscDragInit.manualTransitionLocation),
    transitionLocationWing(viscDragInit.transitionLocationWing),
    transitionLocationBackWing(viscDragInit.transitionLocationBackWing),
    transitionLocationVTP(viscDragInit.transitionLocationVTP),
    doCalibration(viscDragInit.doCalibration),
    CDvisc_CDsum(viscDragInit.CDvisc_CDsum),
    CDvisc_CLfact(viscDragInit.CDvisc_CLfact),
    CDvisc_CLexp(viscDragInit.CDvisc_CLexp),
    CDvisc_CDsum_lowMa(viscDragInit.CDvisc_CDsum_lowMa),
    CDvisc_CLfact_lowMa(viscDragInit.CDvisc_CLfact_lowMa),
    CDvisc_CLexp_lowMa(viscDragInit.CDvisc_CLexp_lowMa),
    jetEngine(viscDragInit.jetEngine),
    Sref_Wing(NAN),
    Swet_Wing(NAN),
    MAC_Wing(NAN),
    avgTtoC_Wing(NAN),
    mXtoC_Wing(NAN),
    eta_ref_Wing(NAN),
    phiAtmXtoC_Wing(NAN),
    QC_Wing(NAN),
    Swet_BackWing(NAN),
    MAC_BackWing(NAN),
    avgTtoC_BackWing(NAN),
    mXtoC_BackWing(NAN),
    eta_ref_BackWing(NAN),
    phiAtmXtoC_BackWing(NAN),
    QC_BackWing(NAN),
    Swet_VTP(NAN),
    MAC_VTP(NAN),
    avgTtoC_VTP(NAN),
    mXtoC_VTP(NAN),
    eta_ref_VTP(NAN),
    phiAtmXtoC_VTP(NAN),
    QC_VTP(NAN),
    Swet_Fuselage(NAN),
    l_Fuselage(NAN),
    maxWidthFuselage(NAN),
    maxHeightFuselage(NAN),
    QC_Fuselage(NAN),
    fuselageUpsweepAngle(NAN),
    Swet_Nacelles(NAN),
    l_Nacelles(NAN),
    maxWidthNacelles(NAN),
    maxHeightNacelles(NAN),
    QC_Nacelles(NAN),
    factorProtuberances(NAN) {
    if (calculateWing) {
        theAircraftComponents.myWing = viscDragInit.theWing;
    }
    if (calculateFuselage) {
        theAircraftComponents.myFuselage = viscDragInit.theFuselage;
    }
    if (calculateNacelles) {
        theAircraftComponents.myNacelles = viscDragInit.theNacelles;
    }
    if (calculateBackWing) {
        theAircraftComponents.myBackWing = viscDragInit.theBackWing;
    }
    if (calculateVTP) {
        theAircraftComponents.myVTP = viscDragInit.theVTP;
    }
    if (!manualTransitionLocation) {
        transitionLocationWing = 0.0;
        transitionLocationBackWing = 0.0;
        transitionLocationVTP = 0.0;
    }
}

void viscDragForTandem::preprocessGeometry() {
    myRuntimeInfo->debug << "viscDragForTandem: preprocess the aircraft geometry" << endl;
    if (calculateWing) {
        Sref_Wing = geom2::measure::reference_area(theAircraftComponents.myWing);
        Swet_Wing = geom2::measure::area(theAircraftComponents.myWing);
        MAC_Wing = geom2::measure::mean_aerodynamic_chord(theAircraftComponents.myWing);
        avgTtoC_Wing = getAverageTtoC(theAircraftComponents.myWing);
        mXtoC_Wing = 0.5;  // Assumption by Raymer for chordwise location of airfoil maximum thickness point for high-speed airfoils
        eta_ref_Wing = 0.65;  // Reference position used for local sweep <- cant  find this value in the source
        phiAtmXtoC_Wing = getLocalSweep(theAircraftComponents.myWing, eta_ref_Wing * getHalfSpan(theAircraftComponents.myWing), mXtoC_Wing, "RADIAN");
        QC_Wing = 1.;  // Assumtion for high, mid, and well fitted low wing
        /*
        // Wetted Surface 
        if (!myAcftPt->theFuselage.empty()) {
            wFuselage = myAcftPt->theFuselage.at(0).getLocalWidth(myAcftPt->theWing.at(i).referencePtx);
            hFuselage = myAcftPt->theFuselage.at(0).getLocalHeight(myAcftPt->theWing.at(i).referencePtx);
            zFuselage = myAcftPt->theFuselage.at(0).getLocalOffset(myAcftPt->theWing.at(i).referencePtx);
        }
        // Check if wing object is located in fuselage
        if (myAcftPt->theWing.at(i).referencePtz < (zFuselage + hFuselage / 2) &&
                myAcftPt->theWing.at(i).referencePtz > (zFuselage - hFuselage / 2)) {  // If tailplane Z-coordinate inside fuselage
            Swet.push_back(myAcftPt->theWing.at(i).getWingSurface() - 2.*(wFuselage * myAcftPt->theWing.at(i).l_i_right));
        } else {
            Swet.push_back(myAcftPt->theWing.at(i).getWingSurface());
        }
        */
    }
    if (calculateBackWing) { //TODO kayra: her türlü tandem icin yap bir tane,  damit nutzer nicht verwirrt wird.
        Swet_BackWing = geom2::measure::area(theAircraftComponents.myBackWing);
        MAC_BackWing = geom2::measure::mean_aerodynamic_chord(theAircraftComponents.myBackWing);
        avgTtoC_BackWing = getAverageTtoC(theAircraftComponents.myBackWing);
        mXtoC_BackWing = 0.4;  // Assumption by Raymer for chordwise location of airfoil maximum thickness point for BackWing
        eta_ref_BackWing = 0.65;  // Reference position used for local sweep
        phiAtmXtoC_BackWing = getLocalSweep(theAircraftComponents.myBackWing, eta_ref_BackWing * getHalfSpan(theAircraftComponents.myBackWing), mXtoC_BackWing, "RADIAN");
        QC_BackWing = 1.04;  // Form factor BackWing
    }
    if (calculateVTP) {
        Swet_VTP = geom2::measure::area(theAircraftComponents.myVTP);
        MAC_VTP = geom2::measure::mean_aerodynamic_chord(theAircraftComponents.myVTP);
        avgTtoC_VTP = getAverageTtoC(theAircraftComponents.myVTP);
        mXtoC_VTP = 0.4;  // Assumption by Raymer for chordwise location of airfoil maximum thickness point for high-speed airfoils
        eta_ref_VTP = 0.65;
        phiAtmXtoC_VTP = getLocalSweep(theAircraftComponents.myVTP, eta_ref_VTP * getHalfSpan(theAircraftComponents.myVTP), mXtoC_VTP, "RADIAN");
        QC_VTP = 1.04;  // Form factor VTP
    }
    if (calculateFuselage) {
        Swet_Fuselage = geom2::measure::area(theAircraftComponents.myFuselage);
        l_Fuselage = geom2::measure::length(theAircraftComponents.myFuselage);
        maxWidthFuselage = geom2::measure::width_max(theAircraftComponents.myFuselage);
        maxHeightFuselage = geom2::measure::height_max(theAircraftComponents.myFuselage);
        QC_Fuselage = 1.0;
        fuselageUpsweepAngle = 3 * PI / 180;  // be creative!!!
    }
    if (calculateNacelles) {
        Swet_Nacelles = getNacelleWettedArea(theAircraftComponents.myNacelles.front());
        l_Nacelles = geom2::measure::length(theAircraftComponents.myNacelles.front());
        maxWidthNacelles = geom2::measure::width_max(theAircraftComponents.myNacelles.front());
        maxHeightNacelles = geom2::measure::height_max(theAircraftComponents.myNacelles.front());
        QC_Nacelles = 1.4;
    }
    // according to [Ray18] (chap. Aerodynamics - Leakage and Protuberances): 2-5% for jet, 5-10% for prop
    if (jetEngine) {
        factorProtuberances = 0.03;
    } else {
        factorProtuberances = 0.07;
    }
    myRuntimeInfo->out << "Viscous drag calculation by Raymer successfully initialized for:" << endl;
    if (calculateWing)
        myRuntimeInfo->out << " -wing" << endl;
    if (calculateBackWing)
        myRuntimeInfo->out << " -BackWing" << endl;
    if (calculateVTP)
        myRuntimeInfo->out << " -VTP" << endl;
    if (calculateFuselage)
        myRuntimeInfo->out << " -fuselage" << endl;
    if (calculateNacelles)
        myRuntimeInfo->out << " -nacelles" << endl;
    if (calculateMisc)
        myRuntimeInfo->out << " -miscellaneous" << endl;
}

std::vector<double> viscDragForTandem::calculateViscDrag(const Polar& aPolar, const double transitionLocationWing,
                                                    const double transitionLocationBackWing, const double transitionLocationVTP) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    if (calculateWing == false) {
        std::stringstream errorMsg;
        errorMsg << "No wing in viscous drag raymer. Unable to calculate drag";
        throwError(__FILE__, __func__, __LINE__, errorMsg.str());
    }
    // vector with component order: 0 = Wing, 1 = BackWing, 2 = VTP, 3 = Fuselage, 4 = Nacelles, 5 = Miscellaneous
    vector <double> theComponentDrags;
    theComponentDrags.push_back(calculateViscDragWing(aPolar, transitionLocationWing));
    // calculate BackWing drag if switch is on
    if (calculateBackWing) {
        theComponentDrags.push_back(calculateViscDragBackWing(aPolar, transitionLocationBackWing));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate VTP drag if switch is on
    if (calculateVTP) {
        theComponentDrags.push_back(calculateViscDragVTP(aPolar, transitionLocationVTP));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate fuselage drag if switch is on
    if (calculateFuselage) {
        theComponentDrags.push_back(calculateViscDragFuselage(aPolar));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate nacelle drag if switch is on
    if (calculateNacelles) {
        theComponentDrags.push_back(calculateViscDragNacelles(aPolar));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate miscellaneous and protuberance drag if switch is on
    if (calculateMisc) {
        theComponentDrags.push_back(calculateViscDragMisc(aPolar));
        // according to [Ray18] (chap. Aerodynamics - Leakage and Protuberances): 2-5% for jet, 5-10% for prop (Page 289)
        double CD_protub = std::accumulate(theComponentDrags.begin(), theComponentDrags.end(), 0.0) * factorProtuberances;
        theComponentDrags.back() += CD_protub;
    } else {
        theComponentDrags.push_back(0.);
    }
    // maybe do a plausibility check here..
    return theComponentDrags;
}

std::vector<double> viscDragForTandem::calculateCorrectedViscDrag(const std::vector<double>& componentDrags, const double aMachNumber, const double CL) {
    /* Drag calibration. First introduced in commit [4eab519c2] based on experimental data in [Lam2010], 23.06.2010. After several changes, 
       the current definition and values in config are from Risse in commit [2ec91580] (02.08.2015) */
    std::vector <double> correctedComponentDrags = componentDrags;
    if (doCalibration) {
        if (aMachNumber > 0.5) {
            correctedComponentDrags.front() += CDvisc_CDsum + CDvisc_CLfact * pow(CL, CDvisc_CLexp);
        } else if (aMachNumber <= 0.5) {
            correctedComponentDrags.front() += CDvisc_CDsum_lowMa + CDvisc_CLfact_lowMa * pow(CL, CDvisc_CLexp_lowMa);
        }
    }
    return correctedComponentDrags;
}

double viscDragForTandem::calculateViscDragWing(const Polar& aPolar, const double transitionLocationWing) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag wing ..." << endl;
    double CDvisc_Wing(0.);
    double reynoldNumberWing = getReynoldsNumber(MAC_Wing, aPolar.altitude, aPolar.MachNumber);
    double Cf;
    double Cflaminar;
    double FF;
    // Friction Coefficient (turbulent) for the wing
    Cf = (getFrictionCoefficient(aPolar.MachNumber, reynoldNumberWing));
    // Friction Coefficient (laminar) for the wing
    Cflaminar = (1.328 / sqrt(reynoldNumberWing));
    // Form factor for the components (eq. 12.30)
    FF = (1. + 0.6 / mXtoC_Wing * avgTtoC_Wing + 100.*pow(avgTtoC_Wing, 4.))
                 * (1.34 * pow(aPolar.MachNumber, 0.18) * pow(cos(phiAtmXtoC_Wing), 0.28));
    // fully turbulent for transitionLocation = 0.0
    CDvisc_Wing = (Cflaminar * transitionLocationWing + Cf * (1. - transitionLocationWing))
                                                            * FF * QC_Wing * Swet_Wing / Sref_Wing;
    return CDvisc_Wing;
}

double viscDragForTandem::calculateViscDragBackWing(const Polar& aPolar, const double transitionLocationBackWing) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag BackWing ..." << endl;
    double CDvisc_BackWing(0.0);
    double reynoldNumberBackWing = getReynoldsNumber(MAC_BackWing, aPolar.altitude, aPolar.MachNumber);
    double Cf;
    double Cflaminar;
    double FF;
    // Friction Coefficient (turbulent) for the BackWing
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldNumberBackWing);
    // Friction Coefficient (laminar) for the BackWing
    Cflaminar = 1.328 / sqrt(reynoldNumberBackWing);
    // Form factor for the BackWing (eq. 12.30)
    FF = ((1. + 0.6 / mXtoC_BackWing * avgTtoC_BackWing + 100.*pow(avgTtoC_BackWing, 4.))
                 * (1.34 * pow(aPolar.MachNumber, 0.18) * pow(cos(phiAtmXtoC_BackWing), 0.28)));
    CDvisc_BackWing = (Cflaminar * transitionLocationBackWing + Cf * (1. - transitionLocationBackWing))
                                                                * FF * QC_BackWing * Swet_BackWing / Sref_Wing;
    return CDvisc_BackWing;
}

double viscDragForTandem::calculateViscDragVTP(const Polar& aPolar, const double transitionLocationVTP) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag VTP ..." << endl;
    double CDvisc_VTP(0.0);
    double reynoldNumberVTP = getReynoldsNumber(MAC_VTP, aPolar.altitude, aPolar.MachNumber);
    double Cf;
    double Cflaminar;
    double FF;
    // Friction Coefficient (turbulent) for component
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldNumberVTP);
    Cflaminar = 1.328 / sqrt(reynoldNumberVTP);
    // Form factor for components
    FF = (1. + 0.6 / mXtoC_VTP * avgTtoC_VTP + 100.*pow(avgTtoC_VTP, 4.))
                 * (1.34 * pow(aPolar.MachNumber, 0.18) * pow(cos(phiAtmXtoC_VTP), 0.28));
    // Store components
    CDvisc_VTP = (Cflaminar * transitionLocationVTP + Cf * (1. - transitionLocationVTP))
                                                            * FF * QC_VTP * Swet_VTP / Sref_Wing;
    return CDvisc_VTP;
}

double viscDragForTandem::calculateViscDragFuselage(const Polar& aPolar) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag Fuselage ..." << endl;
    double CDvisc_Fuselage(0.0);
    double reynoldsNumberFuselage = getReynoldsNumber(l_Fuselage, aPolar.altitude, aPolar.MachNumber);
    double Cf;
    double FF;
    // Friction Coefficient (turbulent) for component
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldsNumberFuselage);
    // Form factor for fuselage
    double f_Fuselage = l_Fuselage / sqrt(maxHeightFuselage * maxWidthFuselage);
    FF = 1. + (60. / pow(f_Fuselage, 3.)) + (f_Fuselage / 400.);
    CDvisc_Fuselage = Cf * FF * QC_Fuselage * Swet_Fuselage / Sref_Wing;
    return CDvisc_Fuselage;
}

double viscDragForTandem::calculateViscDragNacelles(const Polar& aPolar) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag Nacelles ..." << endl;
    double CDvisc_Nacelles(0.0);
    double reynoldsNumberNacelles = getReynoldsNumber(l_Nacelles, aPolar.altitude, aPolar.MachNumber);
    double Cf;
    double FF;
    // Reynolds number of component
    reynoldsNumberNacelles = getReynoldsNumber(l_Nacelles, aPolar.altitude, aPolar.MachNumber);
    // Friction Coefficient (turbulent) for nacelles
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldsNumberNacelles);
    // Form factor for components
    double f_Nacelles = l_Nacelles / sqrt(maxHeightNacelles * maxWidthNacelles);
    FF = 1 + (0.35 / f_Nacelles);
    // Store components
    CDvisc_Nacelles = theAircraftComponents.myNacelles.size() * Cf * FF * QC_Nacelles * Swet_Nacelles / Sref_Wing;
    return CDvisc_Nacelles;
}

double viscDragForTandem::calculateViscDragMisc(const Polar& aPolar) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag Miscellaneous ..." << endl;
    double CDvisc_misc = 0.0;
    if (calculateFuselage) {
        CDvisc_misc = 3.83 * pow(fuselageUpsweepAngle, 2.5) * (PI * 0.5 * maxWidthFuselage * 0.5 * maxHeightFuselage) / Sref_Wing;
    }
    return CDvisc_misc;
}

double viscDragForTandem::getFrictionCoefficient(double flightMach, double ReynoldsNumber) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    // Friction Coefficient (turbulent) for dcomponent
    return 0.455 / (pow(log10(ReynoldsNumber), 2.58) * pow(1 + 0.144 * pow(flightMach, 2.), 0.65));
}
