/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CALCULATEPOLAR_SRC_TANDEM_TANDEMDEFAULTDATA_H_
#define CALCULATEPOLAR_SRC_TANDEM_TANDEMDEFAULTDATA_H_

#include <atmosphere/atmosphere.h>
#include <standardFiles/typedefs.h>
#include <aixml/endnode.h>

#include <moduleBasics/runtimeIO.h>

#include <string>
#include <unordered_map>
#include <vector>
#include <map>

/** \brief Class contains all properties of calculated aircraft objects
*/
class tandemDefaultData {
 public:
    /* Typedefs */
    /** \brief Constructor of aircraft class
     *  \param &mySettings Reference to calculatePolarSettings object
     */
    explicit tandemDefaultData(const std::shared_ptr<RuntimeIO>& rtIO);
    /** \brief Virtual destructor of aircraft class
     */
    virtual ~tandemDefaultData() {}

    double initialCruiseMachNumber;
    std::string engineType;
    /* !!! WARNING: All EndnodeReadOnly Values are converted automatically to SI-Units/RADIAN when read in */
    double bestCruiseCL;
    double designCoGX;
    double designCoGY;
    double designCoGZ;
    double forwardCoGX;
    double forwardCoGY;
    double forwardCoGZ;
    double aftCoGX;
    double aftCoGY;
    double aftCoGZ;
    Point designCoG;              /**< Center of gravity coordinates */
    Point forwardCoG;       /**< Forward position of center of gravity */
    Point aftCoG;           /**< Average flight time position of center of gravity */
    Point cruiseTrimRefCoG; /**< Cruise reference center of gravity position */
    std::unordered_map <std::string, Point> CoGPositions;    /**< Map connecting CoG names with points */
    size_t numberDevices;
    size_t back_numberDevices;
    struct wingDevice {
        std::string type;
        double maxPositiveDeflection;
        double maxNegativeDeflection;
        double innerRelSpanPosition;
        double innerRelFwdChordPosition;
        double innerRelAftChordPosition;
        double outerRelSpanPosition;
        double outerRelFwdChordPosition;
        double outerRelAftChordPosition;
    };
    std::vector<wingDevice> wingDevices;
    std::vector<wingDevice> back_wingDevices;
    Endnode<double>span;
    Endnode<double>back_span;
    Endnode<double>MAC;
    Endnode<double>back_MAC;
    Endnode<double>S_ref;
    Endnode<double>back_S_ref;
    Endnode<double>neutral_point_xCoordinate;
    Endnode<double>C_L_optimumCruise;
    Endnode<double>CL_max_landing;
    Endnode<double>C_LmaxTO;
    Endnode<double>C_LgroundRoll;
    Endnode<std::string>polar_file;
    Endnode<double>max_spoiler_factor;
    Endnode<double>back_max_spoiler_factor;
    void updateNodes(std::string mode);
 private:
    const std::shared_ptr<RuntimeIO>& rtIO_;                    /**< Pointer to the RuntimeIO */
};

#endif  // CALCULATEPOLAR_SRC_TANDEM_TANDEMDEFAULTDATA_H_
