/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CALCULATEPOLAR_SRC_TANDEM_TANDEMDEFAULTSTRATEGY_H_
#define CALCULATEPOLAR_SRC_TANDEM_TANDEMDEFAULTSTRATEGY_H_

#include <moduleBasics/runtimeIO.h>

#include <string>
#include <unordered_map>
#include <vector>
#include <map>
#include <memory>

#include "tandemCalculatePolarConfig.h"
#include "tandemDefaultData.h"
#include "../polar.h"
#include "../methods/liftingLineForTAW.h"
#include "../methods/liftingLinePolar.h"

#include "../methods/viscDragForTandem.h"
#include "../methods/waveDragMason.h"
#include "../methods/semiEmpiricalHighLiftAdaption.h"

class tandemDefaultStrategy {
 public:
    // geometry bools
    // containers for the aircraft geometry
    std::vector <geom2::MultisectionSurface<geom2::AirfoilSection>> liftingSurfaces;    /**< vector which stores all lifting surface geometries (wings and HTP)*/
    geom2::MultisectionSurface<geom2::AirfoilSection> VTP;                              /**< the VTP geometry */
    geom2::MultisectionSurface<geom2::PolygonSection> fuselage;                         /**< the fuselage geometry */
    std::vector <geom2::MultisectionSurface<geom2::PolygonSection>> nacelles;           /**< vector which stores the nacelle geometries*/
    struct wingDevice {
        std::string type;
        double maxPositiveDeflection;
        double maxNegativeDeflection;
        double innerRelSpanPosition;
        double innerRelFwdChordPosition;
        double innerRelAftChordPosition;
        double outerRelSpanPosition;
        double outerRelFwdChordPosition;
        double outerRelAftChordPosition;
    };
    // container for polar data
    struct PolarData {                                                                  /**< struct for handling of the polar data */
        std::unordered_map<size_t, tandemDefaultPolar> cleanPolars;                        /**< clean polars indexed by MachID */
        std::unordered_map<std::string, tandemDefaultPolar> highLiftPolars;                /**< high lift polars indexed by config */
    };
    PolarData currentPolarSet;                                                          /**< temporary polar set which gets overwritten for each new calculation run */
    std::unordered_map<size_t, PolarData> polarDataTrimMap;                             /**< Polar data ordered by stabilizer angle index */
    std::unordered_map<size_t, std::unordered_map<size_t, PolarData>> polarDataCGMap;   /**< Polar data ordered by center of gravity index (0: design, 1: forward, 2; aft) */
                                                                                                                // and by stabilizer angle index
    std::vector<std::string> CoGnames;
    std::vector<std::string> HLconfigs;
    std::string polarXMLFilePath;          /**< Path to the polar XML output file, derived from the I/O file path */
    std::string designPolarXMLFilePath;
    std::string LILI_Version;
    // variables for the acftXML output
    double C_LmaxLanding;
    double C_LmaxTO;
    double C_LoptimumCruise;

    void run();
    /** \brief Function initializes liftingLinePolar with the required variables
     *  \param myLiLiInit the lifting line polar initializer method with gets initialized
     */
    void initializeLiLiPolar(LiLiPolarInitialization *myLiLiInit);
    /** \brief Function initializes viscDragRaymer with the required variables
     *  \param myViscDragRaymer the viscous drag raymer method with gets initialized
     */
    void initializeViscDragForTandem(viscDragForTandemInitialization *myViscDragInit);
    /** \brief Function initializes waveDragMason with the required variables
     *  \param myWaveDragMason the wave drag mason method with gets initialized
     */
    void initializeWaveDragMason(waveDragInitialization *myWaveDragInit);
    /** \brief Function with builds and saves the entire aircraft geometry
     *  \param stabAngle the angle of the stabilizer relative to the design angle, which gets altered for the trimm setting [deg]
     */
    void createAircraftGeometry(const double backWingAngle);
    /** \brief Function witch executes all aero calculation methods
     *  \param CoGPosition the 3D center of gravity coordinate as reference for the aero moments [m]
     */    
    void runAeroCalculations(const Point CoGPosition, const double currentIStabAngle);
    /** \brief Function to copy the polar set from liftingLinePolar to the strategyClass
     *  \param liftingLinePolar the lifting line for TAW method, which generated the clean polar set
     */
    void getPolarDataFromLiftingLine(const liftingLinePolar& liftingLinePolar);
    /** \brief Function to calculate the viscous drag for the current polar set
     *  \param myViscDragForTandem the viscous drag raymer method to calculate the drag
     */
    void calculateViscDragForTandem(viscDragForTandem *myViscDragForTandem);
    /** \brief Function to calculate the wave drag for the current polar set
     *  \param myWaveDragMason the wave drag mason method to calculate the drag
     */
    void calculateWaveDragMason(waveDragMason *myWaveDragMason);
    void sumUpDragComponents();
    void copyWingDevicesAndStabilizer(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption);
    void copyCleanPolarForHLCalculation(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, liftingLinePolar *myLiftingLinePolar);
    void calculateHighLiftPolars(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, const double CoG_X_Position, const double currentIStabAngle);
    /** \brief Copy the temporary current polar set to the permanent polarDataCGMap
     *  \param CGID index of the current center of gravity position
     *  \param stabID index of the current stabilizer angle
     */
    void copyPolarsToCGMap(const size_t CGID, const size_t stabID);
    /** \brief Function to delete temporary data befor next aero calculation run
     */
    void clearDataForNextCalculationRun();
    /** \brief Function calls and handles the trim interpolation calculations and data transfer
     */
    void doTrimInterpolation();
    /** \brief Function calls and handles the high lift trim interpolation calculations and data transfer
     */
    void doTrimInterpolationHighLift();
    /** \brief Function copies the clean polars from the temporary current polar set to the permanent trimmed polar data map
     */
    void copyCleanPolars();
    /** \brief Function copies the high lift polars from the temporary current polar set to the permanent trimmed polar data map
     */
    void copyHighLiftPolars();

    void calibratePolars();

    void calibrateCleanPolars();

    void calibrateHighLiftPolars();

    void writePolarXML();

    void writePolarData();

    void writeViscDragData();

    double getCLoptCruise();

    double getMaxCL(const tandemDefaultPolar& aPolar);

    void setReferenceValues();

    void setOutputValues();

    void update_xml_entries(std::string mode);


    tandemDefaultStrategy(const std::shared_ptr<RuntimeIO>& rtIO, const std::shared_ptr<tandemCalculatePolarConfig>& config);
    virtual ~tandemDefaultStrategy() {}

 private:
    std::shared_ptr<node> aircraft_xml;
    const std::shared_ptr<RuntimeIO>& rtIO_;                    /**< Pointer to the RuntimeIO */
    const std::shared_ptr<tandemDefaultData> data_;        /**< Pointer to the aircraft data */
    const std::shared_ptr<tandemCalculatePolarConfig>& config_;    /**< Pointer to the config data */
};
#endif  // CALCULATEPOLAR_SRC_TANDEM_TANDEMDEFAULTSTRATEGY_H_
