/**
 * \file cantileverTandemWingDesign.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief
 * \version 0.1
 * \date 2023-12-05
 *
 * @copyright Copyright (c) 2023
 *
 */

#include "cantileverTandemWingDesign.h"

#include <aircraftGeometry2/io/dat.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <atmosphere/atmosphere.h>
#include <unitConversion/unitConversion.h>
#include <CGAL/Simple_cartesian.h>
#include <CGAL/Surface_mesh.h>

#include <algorithm>
#include <cmath>
#include <map>
#include <tuple>

#include "lib/design_methods/aspectRatioMethods.h"
#include "lib/design_methods/controlDeviceMethods.h"
#include "lib/design_methods/dihedralMethods.h"
#include "lib/design_methods/relativeKinkPositionMethods.h"
#include "lib/design_methods/sparMethods.h"
#include "lib/design_methods/sweepAngleMethods.h"
#include "lib/design_methods/taperRatioMethods.h"
#include "lib/design_methods/wingAreaMethods.h"
#include "lib/design_methods/wingProfileMethods.h"
#include "lib/general_methods/icao_faa_codes.h"
#include "lib/general_methods/newtonAlgorithm.h"


geom2::AirfoilSection get_rescaled_airfoil_geometry(const std::string& airfoildatadir,
                                                    std::tuple<double, double, std::string> wing_profile_and_thickness); 

namespace tandem {
namespace cantilever {
Wing::Wing(const std::shared_ptr<RuntimeIO>& rtIO)
    : rtIO(rtIO),
      config(std::make_shared<tandem::cantilever::WingDesignConfig>()), 
      data(std::make_shared<tandem::cantilever::WingDesignData>()),
      reporter(rtIO) {
  /* Setup design modes */
  design_mode_runner["mode_0"] = [this]() { standard_design(); };

  /* Setup mass modes */
  mass_mode_runner["mode_0"] = [this]() { flops_mass(); };    //TODO kayra: buradan ayari
  mass_mode_runner["mode_1"] = [this]() { chiozzotto_wer_mass(); };

}

void Wing::initialize() {
  myRuntimeInfo->out << "Initializing wing design ..kayra. [START]" << std::endl;
  /* Create Airfoil directory in project path */ 
  rtIO->createAirfoilDataDir();

  /* Read config */
  config->read(rtIO->moduleConfig);

  /* Read data */
  data->read(rtIO->acxml);

  /* fuselage must be read seperately since path to geometry needed */
  std::filesystem::path fuselage_path = rtIO->getGeometryDir();
  data->read_fuselage(rtIO->acxmlAccess, fuselage_path);

  /* Store method to variable */
  selected_design_mode = config->wing_design_mode.value();
  selected_mass_mode = config->wing_mass_mode.value();


  airfoils_library = std::make_shared<Airfoils>(static_cast<std::filesystem::path>(config->common_airfoil_data_path.value()));
  airfoils_library->add_directory_airfoils(rtIO->getAirfoilDataDir());

  myRuntimeInfo->out << "Initializing wing design ... [FINISHED]" << std::endl;
}
void Wing::run() {
  myRuntimeInfo->out << "Run wing design ... [START]" << std::endl;

  myRuntimeInfo->out << "Selected design method ... [START]" << std::endl;
  design_mode_runner[selected_design_mode]();
  myRuntimeInfo->out << "Selected design method ... [FINISHED]" << std::endl;

  myRuntimeInfo->out << "Selected mass method   ... [START]" << std::endl;
  mass_mode_runner[selected_mass_mode]();    
  myRuntimeInfo->out << "Selected mass method   ... [FINISHED]" << std::endl;

  myRuntimeInfo->out << "Run wing design ... [FINISHED]" << std::endl;
}
void Wing::update() {
  myRuntimeInfo->out << "Update designed wing ... [START]" << std::endl;
  data->update(rtIO->acxml);
  myRuntimeInfo->out << "Update designed wing ... [FINISHED]" << std::endl;
}
void Wing::report() {
  myRuntimeInfo->out << "Report designed wing ... [START]" << std::endl;
  /* Sets body for html report */
  set_html_body();

  /* Generate reports */
  reporter.generateReports();
  /* Generate wing stl */
  geom2::Mesh wing_mesh = geom2::transform::to_mesh(data->wing);
  CGAL::IO::write_STL("front_wing_left.stl", wing_mesh);
  geom2::Mesh back_wing_mesh = geom2::transform::to_mesh(data->back_wing);
  CGAL::IO::write_STL("back_wing_left.stl", back_wing_mesh);
  geom2::Mesh combined_mesh = combine_meshes(wing_mesh, back_wing_mesh);
  CGAL::IO::write_STL("front_and_back_wing_left.stl", combined_mesh);
  myRuntimeInfo->out << "Report designed wing ... [FINISHED]" << std::endl;
}
void Wing::save() {
  // nothing to do
}

/* == Design ========================================================================================================
 */

void Wing::standard_design() {
  /* Run standard method */
  myRuntimeInfo->out << "Wing Design ... [Start]" << std::endl;

  double f_L = geom2::measure::length(data->fuselage);
  myRuntimeInfo->out << "fuselage length: " << f_L << std::endl;

  myRuntimeInfo->out << "relative longitudinal position of tandem wing set and front wing: " <<  data->specification_longitudinal_position.value() << " of fuselage length" << std::endl;
  myRuntimeInfo->out << "relative LE stagger: " << config->user_relative_LE_stagger.value() << " of fuselage length" << std::endl;
  myRuntimeInfo->out << "relative position of rear wing: " << data->specification_longitudinal_position.value() + config->user_relative_LE_stagger.value() << " of fuselage length" << std::endl;

  //TODO kayra: bunu dah altta chord length ile birlikte cagirsam mantikli olabilir. ama wing gemometry calculationa da dikkat ederek cagirmak lazim olur.
  if (data->specification_longitudinal_position.value() + config->user_relative_LE_stagger.value() > 1) {
      throwError(__FILE__,__func__,__LINE__,"Error - Position of rear wing exceeds fuselage length. Please decrease your tandem wing set position value or your stagger value");
  }

  /* check seperately if wings are kinked */  //TODO kayra: kink olayini user defined yapabilirsin tamamen.
  bool is_wing_kinked = is_wing_kinked_check();
  bool is_back_wing_kinked = is_back_wing_kinked_check();

  /* get maximum thickness to chord ratios of wings */
  double max_thickness_to_chord_ratio = get_max_thickness_to_chord_ratio();
  myRuntimeInfo->out << "max_thickness_to_chord_ratio: " << max_thickness_to_chord_ratio << std::endl;
  double back_max_thickness_to_chord_ratio = get_back_max_thickness_to_chord_ratio();
  myRuntimeInfo->out << "back_max_thickness_to_chord_ratio: " << back_max_thickness_to_chord_ratio << std::endl;


  /* get mtom */
  double maximum_takeoff_mass = get_maximum_takeoff_mass();
   myRuntimeInfo->out << "MTOM: " << maximum_takeoff_mass << std::endl;

  // Compute total wing area based on selected mode
  myRuntimeInfo->out << "Calculated wing area is the TOTAL WING AREA of both wings!" << std::endl;
  const double installed_total_wing_area =
      get_design_wing_area(config->total_wing_area_calculation_mode.value(), config->user_total_wing_area.value(),
                           maximum_takeoff_mass, data->sizing_point_wing_loading.value());


  const double wing_area_ratio = config->user_wing_area_ratio.value();
  
  myRuntimeInfo->out <<"Wing Area Ratio: "<< wing_area_ratio << std::endl;


  const double installed_front_wing_area = installed_total_wing_area / (wing_area_ratio + 1);
  const double installed_back_wing_area = installed_total_wing_area - installed_front_wing_area;

  myRuntimeInfo->out <<"S_1: "<< installed_front_wing_area << std::endl;
  myRuntimeInfo->out <<"S_2: "<< installed_back_wing_area << std::endl;


  // Compute front wing quarter chord based on selected mode            
  myRuntimeInfo->out << "1/4 sweep of front Wing: " << std::endl;
  const double installed_quarter_chord_sweep = get_wing_quarter_chord_sweep(
      config->sweep_calculation_mode.value(), config->user_sweep_angle_at_quarter_chord.value(),
      data->specification_design_mach.value(), data->sizing_point_wing_loading.value(),
      data->specification_design_altitude.value(), max_thickness_to_chord_ratio,
      config->delta_drag_divergence_to_mach_design.value(), config->korn_technology_factor.value()); 
 
  myRuntimeInfo->out << "1/4 sweep of rear Wing: " << std::endl;    
  // Compute back wing quarter chord based on selected mode          
  const double installed_back_quarter_chord_sweep = get_wing_quarter_chord_sweep(
      config->sweep_calculation_mode.value(), config->user_back_sweep_angle_at_quarter_chord.value(),
      data->specification_design_mach.value(), data->sizing_point_wing_loading.value(),
      data->specification_design_altitude.value(), back_max_thickness_to_chord_ratio,
      config->back_delta_drag_divergence_to_mach_design.value(), config->back_korn_technology_factor.value());


  // Comptue aspect ratios and wing spans
  std::map<std::string, std::tuple<double, double>> limits_field_length_and_wing_span =
      icao_easa_aerodrome_reference_code(data->specification_icao_aerodrome_reference_code.value());

  /* Current limits of wing span according to reference code data*/   //TODO kayra: lower limit neden var
  const auto [lower_wing_span, upper_wing_span] = limits_field_length_and_wing_span["wing_span"];

  /* Compute aspect ratios based on selected modes*/
  //TODO kayra spanden gidip AR bulman gerekebilir. uzun installed tapera kadar.ayrica sadece user a izin verebilirsin biri icin
  //TODO kayra ayrica pitch up limit function icin Cop iki kanatta da ileri gidiyor ama bir finetuning gerekli mi bilmiyorum
  //TODO kayra: aspect ratiolari da sinirlandiracak bir fkt yazabilirsin user birbirinden alakasiz vermesin diye yada ARR ile yap.
  //myRuntimeInfo->out <<"WSR: "<< wing_span_ratio << std::endl;//TODO kayra:suan yanlislik yaratiyor
  //TODO kayra: bu design icin gereksiz bir parametre olarak kalabilir eger farkli yapmazsak.

  myRuntimeInfo->out <<"initial AR_1 of front wing: "<< std::endl;          
  double aspect_ratio = get_design_aspect_ratio(config->aspect_ratio_calculation_mode.value(),
                                                config->user_aspect_ratio.value(), installed_quarter_chord_sweep);
                                                
  myRuntimeInfo->out <<"initial AR_2 of back wing: "<< std::endl; 
  double back_aspect_ratio = get_design_aspect_ratio(config->aspect_ratio_calculation_mode.value(),
                                                config->user_back_aspect_ratio.value(), installed_back_quarter_chord_sweep);
  
  /* Determine spans and possible aspect ratio limits */
  double span = sqrt(aspect_ratio * installed_front_wing_area);
  myRuntimeInfo->out <<"initial calculated b_1 for front wing: "<< span << std::endl;
  double back_span = sqrt(back_aspect_ratio * installed_back_wing_area);
  myRuntimeInfo->out <<"initial calculated b_2 for back wing: "<< back_span << std::endl;


  double maximum_possible_aspect_ratio = pow(upper_wing_span, 2.) / installed_front_wing_area;
  double minimum_possible_aspect_ratio = pow(lower_wing_span, 2.) / installed_front_wing_area;
  double maximum_possible_back_aspect_ratio = pow(upper_wing_span, 2.) / installed_back_wing_area;
  double minimum_possible_back_aspect_ratio = pow(lower_wing_span, 2.) / installed_back_wing_area;

  /* if current aspect ratios exceed maximum possible aspect ratio -> set aspect ratios to maximum possible aspect ratio
   * and set span to according wing span */
  if (aspect_ratio > maximum_possible_aspect_ratio) {
    myRuntimeInfo->warn << "Initial aspect ratio for front wing > maximum_possible_aspect_ratio for front wing" << std::endl;
    myRuntimeInfo->warn << "Switch to maximum possible aspect_ratio" << std::endl;
    aspect_ratio = maximum_possible_aspect_ratio;
    span = upper_wing_span;
  }
  if (back_aspect_ratio > maximum_possible_back_aspect_ratio) {
    myRuntimeInfo->warn << "Initial aspect ratio for back wing > maximum_possible_back_aspect_ratio for back wing" << std::endl;
    myRuntimeInfo->warn << "Switch to maximum possible back_aspect_ratio" << std::endl;
    back_aspect_ratio = maximum_possible_back_aspect_ratio;
    back_span = upper_wing_span;
  }
  // Check if current aspect ratios are below lower span limits -> set aspect ratios to minimum possible aspect ratios and
  // set spans to according wing spans */
  if (aspect_ratio < minimum_possible_aspect_ratio) {
    myRuntimeInfo->warn << "Current aspect ratio for front wing breach lower span limit ..." << std::endl;
    myRuntimeInfo->warn << "Switch to minimum possible aspect_ratio" << std::endl;
    aspect_ratio = minimum_possible_aspect_ratio;
    span = lower_wing_span;
  }
  if (back_aspect_ratio < minimum_possible_back_aspect_ratio) {
    myRuntimeInfo->warn << "Current aspect ratio for back wing breach lower span limit ..." << std::endl;
    myRuntimeInfo->warn << "Switch to minimum possible back_aspect_ratio" << std::endl;
    back_aspect_ratio = minimum_possible_back_aspect_ratio;
    back_span = lower_wing_span;
  }

  /* Freeze aspect ratios and spans */
  const double installed_aspect_ratio = aspect_ratio;
  const double installed_back_aspect_ratio = back_aspect_ratio;
  const double installed_span = span;
  const double installed_back_span = back_span;
  
  myRuntimeInfo->out <<"Corrected final values for aspect ratios and wing spans according to the airport limits: "<< std::endl;
  myRuntimeInfo->out <<"AR_1: "<< installed_aspect_ratio << std::endl;
  myRuntimeInfo->out <<"AR_2: "<< installed_back_aspect_ratio << std::endl;
  myRuntimeInfo->out <<"b_1: "<< installed_span << std::endl;
  myRuntimeInfo->out <<"b_2: "<< installed_back_span << std::endl;

  const double installed_wing_span_ratio = installed_back_span / installed_span; //TODO kayra: just to show
  myRuntimeInfo->out <<"Wing Span Ratio: "<< installed_wing_span_ratio << std::endl;

  const double installed_aspect_ratio_ratio = installed_back_aspect_ratio / installed_aspect_ratio; //TODO kayra: just to show
  myRuntimeInfo->out <<"Aspect Ratio Ratio: "<< installed_aspect_ratio_ratio << std::endl;

  if(installed_wing_span_ratio > 1.2 || installed_wing_span_ratio < 0.83 || 
      installed_aspect_ratio_ratio  > 2|| installed_aspect_ratio_ratio < 0.83) {
        myRuntimeInfo->warn << "Please consider that generally for a properly designed tandem wing configuration Wing Area Ratio, Wing Span Ratio and Aspect Ratio Ratio should be between 0.83 and 1.2" << std::endl;
  }

  // Compute taper ratios based on selected modes
  myRuntimeInfo->out <<"Taper Ratio of forward wing: "<< std::endl;
  const double installed_taper_ratio =
      get_design_taper_ratio(config->taper_ratio_calculation_mode.value(), config->user_taper_ratio.value(),
                             installed_aspect_ratio, installed_quarter_chord_sweep);
  myRuntimeInfo->out <<"Taper Ratio of rear wing: "<< std::endl;
  const double installed_back_taper_ratio =
      get_design_taper_ratio(config->taper_ratio_calculation_mode.value(), config->user_back_taper_ratio.value(),
                             installed_back_aspect_ratio, installed_back_quarter_chord_sweep);

  // Comptue dihedrals based on selected modes 
  //TODO kayra: paper var buna göre yap
  myRuntimeInfo->out <<"Dihedral Angle of forward wing: "<< std::endl;
  const double installed_dihedral = get_design_dihedral(
      config->dihedral_calculation_mode.value(), data->specification_wing_mounting.value(),
      config->user_dihedral.value(), installed_quarter_chord_sweep, data->specification_design_mach.value());
  myRuntimeInfo->out <<"Dihedral Angle of rear wing: "<< std::endl;
  const double installed_back_dihedral = get_design_dihedral(
      config->dihedral_calculation_mode.value(), data->specification_back_wing_mounting.value(),
      config->user_back_dihedral.value(), installed_back_quarter_chord_sweep, data->specification_design_mach.value());

  // Generate wings
  myRuntimeInfo->out <<"FORWARD WING PROFILE: "<< std::endl;
  if (is_wing_kinked) {
    data->wing =
        calculate_kinked_wing_geometry(installed_taper_ratio, installed_aspect_ratio, installed_quarter_chord_sweep,
                                       installed_dihedral, installed_front_wing_area, installed_span);
  } else {
    data->wing =
        calculate_unkinked_wing_geometry(installed_taper_ratio, installed_aspect_ratio, installed_quarter_chord_sweep,
                                         installed_dihedral, installed_front_wing_area, installed_span);
  }
  myRuntimeInfo->out <<"REAR WING PROFILE: "<< std::endl;
  if (is_back_wing_kinked) {
    data->back_wing =
        calculate_kinked_back_wing_geometry(installed_back_taper_ratio, installed_back_aspect_ratio, installed_back_quarter_chord_sweep,
                                       installed_back_dihedral, installed_back_wing_area, installed_back_span);
  } else {
    data->back_wing =
        calculate_unkinked_back_wing_geometry(installed_back_taper_ratio, installed_back_aspect_ratio, installed_back_quarter_chord_sweep,
                                         installed_back_dihedral, installed_back_wing_area, installed_back_span);
  }

  // Set wing names to constant names
  data->wing.name = "FRONT WING";
  data->back_wing.name ="BACK WING";

  data->wing_position.set_xyz(calculate_wing_position());

  myRuntimeInfo->out << "front wing position" << " X:" <<data->wing_position.x.value()<<" Y:"<<data->wing_position.y.value()<<" Z:"<<data->wing_position.z.value()<< std::endl;
  
  data->back_wing.origin = calculate_back_wing_position();

  myRuntimeInfo->out << "back wing position" << " X:" <<data->back_wing.origin.x()+data->wing_position.x.value()<<" Y:"<<data->back_wing.origin.y()+data->wing_position.y.value()<<" Z:"<<data->back_wing.origin.z()+data->wing_position.z.value()<< std::endl;


  // Set spar devices 
  data->spars = get_spars(config->spars_mode.value(), config->user_defined_spars);

  data->back_spars = get_spars(config->back_spars_mode.value(), config->user_defined_back_spars);
  // Set control devices
  
  double eta_kink = 0.;
  double back_eta_kink = 0.;

  if (is_wing_kinked_check()) {
    eta_kink = get_design_relative_kink_position(
        config->relative_kink_position_calculation_mode.value(), config->user_relative_kink.value(),
        config->track_based_initial_relative_kink.value(), data->track_based_relative_kink.value());
  }
    if (is_back_wing_kinked_check()) {
    back_eta_kink = get_design_relative_kink_position(
        config->back_relative_kink_position_calculation_mode.value(), config->user_back_relative_kink.value(),
        config->track_based_back_initial_relative_kink.value(), data->track_based_back_relative_kink.value());
  }

  auto [devices, deflections] = get_control_devices(config->control_device_mode.value(), config->user_defined_control_devices,
                                              data->spars, data->wing, data->fuselage, is_wing_kinked_check(), eta_kink,
                                              config->high_lift_device_type_leading_edge.value(),
                                              config->high_lift_device_type_trailing_edge.value());

  auto [back_devices, back_deflections] = get_control_devices(config->back_control_device_mode.value(), config->user_defined_back_control_devices,
                                              data->back_spars, data->back_wing, data->fuselage, is_back_wing_kinked_check(), back_eta_kink,
                                              config->back_high_lift_device_type_leading_edge.value(),
                                              config->back_high_lift_device_type_trailing_edge.value());

  data->control_devices = devices;
  data->control_devices_deflections = deflections;
  data->back_control_devices = back_devices;
  data->back_control_devices_deflections = back_deflections;

  /* Check control device boundaries - if inside fuselage */
  if ((config->control_device_mode.value() == "mode_0") &&
      (rtIO->aircraft_configuration_type() == "tandem")) {
    double eta_fuselage_max = geom2::measure::width_max(data->fuselage) / geom2::measure::span(data->wing); //TODO kayra: ok olabilir bu hesap devices uzak olsun
    for (auto& device : data->control_devices) {
      size_t sec_id = 0;
      for (auto& section : device.sections) {
        if (section.origin.z() < eta_fuselage_max * 1.1) {
          myRuntimeInfo->warn << std::format(
              "User defined control device for forward wing {} (section {}) inside fuselage width - set to span position to 1.1 "
              "fuselage width\n",
              device.name, sec_id);
          section.origin = geom2::Point_3(section.origin.x(), section.origin.y(), eta_fuselage_max * 1.1);
        }
        sec_id++;
      }
    }
  }
  if ((config->back_control_device_mode.value() == "mode_0") &&
      (rtIO->aircraft_configuration_type() == "tandem")) {
    double back_eta_fuselage_max = geom2::measure::width_max(data->fuselage) / geom2::measure::span(data->back_wing);
    for (auto& device : data->back_control_devices) {
      size_t sec_id = 0;
      for (auto& section : device.sections) {
        if (section.origin.z() < back_eta_fuselage_max * 1.1) {
          myRuntimeInfo->warn << std::format(
              "User defined control device for rear wing {} (section {}) inside fuselage width - set to span position to 1.1 "
              "fuselage width\n",
              device.name, sec_id);
          section.origin = geom2::Point_3(section.origin.x(), section.origin.y(), back_eta_fuselage_max * 1.1);
        }
        sec_id++;
      }
    }
  }
}

bool Wing::is_wing_kinked_check() {
  myRuntimeInfo->out << "Front Wing kink check ..." << std::endl;

  /* Check if front wing mounting is low &  landing gear mounted on front wing */
  if (data->specification_wing_mounting.value()== "low" &&
      data->specification_landing_gear_mounting.value() == "wing_mounted") {
    myRuntimeInfo->out << "Front Wing is kinked  ... [TRUE]" << std::endl;
    return true;
  } else {
    myRuntimeInfo->out << "Front Wing is kinked  ... [FALSE]" << std::endl;
    return false;
  }
}

bool Wing::is_back_wing_kinked_check() {
  myRuntimeInfo->out << "Back Wing kink check ..." << std::endl;   

  /* Check if back wing mounting is low &  landing gear mounted on back wing */
  if (data->specification_back_wing_mounting.value()== "low" &&
      data->specification_landing_gear_mounting.value() == "tandem_back_wing_mounted") { 
    myRuntimeInfo->out << "Back Wing is kinked  ... [TRUE]" << std::endl;
    return true;
  } else {
    myRuntimeInfo->out << "Back Wing is kinked  ... [FALSE]" << std::endl;
    return false;
  }
}

double Wing::get_max_thickness_to_chord_ratio() {
  // Get maximum thickness to chord ratio based on wing profile and thickness mode
  // If user - get maximum thickness from elements
  // Else use max_thickness to chord ratio from sweep calculation method
  if (config->wing_profile_and_thickness_calculation_mode.value() == "mode_0") {
    std::vector<double> to_check;
    for (auto item : config->user_thickness_to_chord) {
      to_check.push_back(item.value());
    }
    return *std::max_element(to_check.begin(), to_check.end());
  } else {
    return config->max_thickness_to_chord_ratio.value();
  }
}

double Wing::get_back_max_thickness_to_chord_ratio() {
  // Get maximum thickness to chord ratio based on back wing profile and thickness mode
  // If user - get maximum thickness from elements
  // Else use max_thickness to chord ratio from sweep calculation method
  if (config->back_wing_profile_and_thickness_calculation_mode.value() == "mode_0") {
    std::vector<double> to_check;
    for (auto item : config->user_back_thickness_to_chord) {
      to_check.push_back(item.value());
    }
    return *std::max_element(to_check.begin(), to_check.end());
  } else {
    return config->back_max_thickness_to_chord_ratio.value();
  }
}

double Wing::get_maximum_takeoff_mass() {
  return data->mtom_mass_properties.data["mass"].value();
}

//TODO kayra: buralarda cok fazla ayarlama yapman lazim simdiden. gereksizleri de silmen lazim. eger arka taraf pos ayarlamasi yaparsan wing geometri hesabina da eklemen lazim.ayrica mass lere de

geom2::Point_3 Wing::calculate_wing_position() {
  // Get cg position - if overall not exists -> use fuselage cg
  double x_wing_apex, x_cg, y;
  std::tuple<double, double> z_limits;

  double y_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  double x_mac_position_quarter_chord_referred_to_centerline =
      geom2::measure::offset_LE(data->wing, y_mac_position).x() +   //TODO kayra: if twist this should be adjusted
      0.25 * geom2::measure::chord(data->wing, y_mac_position);

  const double fuselage_length = geom2::measure::length(data->fuselage);
  const double fuselage_maximum_width = geom2::measure::width_max(data->fuselage);
  const double fuselage_maximum_height = geom2::measure::height_max(data->fuselage); //TODO kayra:daha ayarini yapmamisim. keal beam de karisiklik yaratabilir.

  double fuselage_upper_z;
  double fuselage_lower_z;

  /* Keel beam height as fraction of fuselage height*/
  double keel_beam_height = 0.005 * fuselage_maximum_height;  // Assumed 0.5% of fuselage height

  /* Define wing position limits */
  if (data->specification_wing_mounting.value() == "low") {
    fuselage_upper_z = data->fuselage.origin.z();
    fuselage_lower_z = data->fuselage.origin.z() - fuselage_maximum_height * 0.5;
    keel_beam_height = 0.005 * (fuselage_upper_z - fuselage_lower_z);
    z_limits = {fuselage_lower_z + keel_beam_height, fuselage_upper_z};
  } else if (data->specification_wing_mounting.value() == "high") {
    z_limits = {fuselage_lower_z, fuselage_upper_z - keel_beam_height};
  } else {
    z_limits = {fuselage_lower_z, fuselage_upper_z};
  }

  // Assume center position
  if (data->mtom_mass_properties.data["x"].value() < 0.01 && !data->fuselage.sections.empty()) {
    x_cg = fuselage_length * 0.45;
  } else {
    x_cg = data->mtom_mass_properties.data["x"].value();
  }
  auto const [lim_low, lim_up] = z_limits;
  double z_position = lim_low + 0.5 * (lim_up - lim_low);
  return geom2::Point_3(data->specification_longitudinal_position.value()*fuselage_length, 0., z_position);
}

geom2::Point_3 Wing::calculate_back_wing_position() {
  // Get cg position - if overall not exists -> use fuselage cg
  double x_wing_apex, x_cg, y;
  std::tuple<double, double> z_limits;

  double y_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->back_wing);
  double x_mac_position_quarter_chord_referred_to_centerline =
      geom2::measure::offset_LE(data->back_wing, y_mac_position).x() + //TODO kayra: if twist this should be adjusted
      0.25 * geom2::measure::chord(data->back_wing, y_mac_position);

  const double fuselage_length = geom2::measure::length(data->fuselage);
  const double fuselage_maximum_width = geom2::measure::width_max(data->fuselage);
  const double fuselage_maximum_height = geom2::measure::height_max(data->fuselage);  //TODO kayra:daha ayarini yapmamisim. keal beam de karisiklik yaratabilir.

  double fuselage_upper_z;
  double fuselage_lower_z;

  /* Keel beam height as fraction of fuselage height*/
  double keel_beam_height = 0.005 * fuselage_maximum_height;  // Assumed 0.5% of fuselage height

  /* Define wing position limits */
  if (data->specification_back_wing_mounting.value() == "low") {
    fuselage_upper_z = data->fuselage.origin.z();
    fuselage_lower_z = data->fuselage.origin.z() - fuselage_maximum_height * 0.5;
    keel_beam_height = 0.005 * (fuselage_upper_z - fuselage_lower_z);
    z_limits = {fuselage_lower_z + keel_beam_height, fuselage_upper_z};
  } else if (data->specification_back_wing_mounting.value() == "high") {
    z_limits = {fuselage_lower_z, fuselage_upper_z - keel_beam_height};
  } else {
    z_limits = {fuselage_lower_z, fuselage_upper_z};
  }

  // Assume center position
  if (data->mtom_mass_properties.data["x"].value() < 0.01 && !data->fuselage.sections.empty()) {
    x_cg = fuselage_length * 0.45;
  } else {
    x_cg = data->mtom_mass_properties.data["x"].value();
  }
  auto const [lim_low, lim_up] = z_limits;
  double z_position = lim_low + 0.5 * (lim_up - lim_low);

  return geom2::Point_3(config->user_relative_LE_stagger.value()*fuselage_length, 0., z_position - data->wing_position.z.value());
}


geom2::MultisectionSurface<geom2::AirfoilSection> Wing::calculate_kinked_wing_geometry(
    const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
    const double wing_area, const double span) {
  double maximum_inner_trailing_edge_sweep = get_design_maximum_inner_trailing_edge_sweep(
      config->relative_kink_position_calculation_mode.value(),
      config->user_max_inner_trailing_edge_sweep.value(),          // old converted to radians
      config->track_based_max_inner_trailing_edge_sweep.value());  // old converted to radians

  const double max_thickness_to_chord_ratio = get_max_thickness_to_chord_ratio();

  // Get normalized half span values - eta
  const double half_span = 0.5 * span;

  const double half_fuselage_width = 
    geom2::measure::width(data->fuselage, geom2::measure::length(data->fuselage) * data->specification_longitudinal_position.value()) * 0.5; //TODO kayra: max olmasi kesin degil wingin. ondan böyle yaptim
  const double eta_fuselage = half_fuselage_width / half_span;
  double eta_centerline = 0. / half_span;
  double eta_tip = half_span / half_span;
  std::vector<double> eta_values = {eta_centerline, eta_fuselage, eta_tip};

  double eta_kink = get_design_relative_kink_position(
      config->relative_kink_position_calculation_mode.value(), config->user_relative_kink.value(),
      config->track_based_initial_relative_kink.value(), data->track_based_relative_kink.value());
  eta_values.push_back(eta_kink);

  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      get_wing_profile_and_thickness(config->wing_profile_and_thickness_calculation_mode.value(),
                                     config->user_wing_profiles_and_thickness_vector,
                                     config->torenbeek_jenkinson_wing_profile.value(), max_thickness_to_chord_ratio,
                                     config->airfoil_critical_factor.value(), data->specification_design_mach.value(),
                                     data->specification_design_altitude.value(), quarter_chord_sweep,
                                     config->delta_drag_divergence_to_mach_design.value(),
                                     data->sizing_point_wing_loading.value(), eta_fuselage, eta_values);

  // Geometry computation functions
  auto condition_trailing_edge_max = [=](const double inner_taper_ratio) {
    return (4. / 3. * (tan(quarter_chord_sweep) - tan(maximum_inner_trailing_edge_sweep)) * half_span *
            (eta_kink - eta_fuselage) / (1 - inner_taper_ratio));
  };

  auto root_chord_function = [=](const double inner_taper_ratio) {
    return wing_area / (2. * half_span *
                        (eta_fuselage + (1 + inner_taper_ratio) * 0.5 * (eta_kink - eta_fuselage) +
                         (inner_taper_ratio + taper_ratio) * 0.5 * (eta_tip - eta_kink)));
  };

  // Maximum possible taper ratio
  const double eta_ratio = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  const double taper_ratio_inner_max = (eta_ratio + taper_ratio) / (1 + eta_ratio);
  auto taper_ratio_inner_searcher = [=](const double inner_taper_ratio) {
    return root_chord_function(inner_taper_ratio) - condition_trailing_edge_max(inner_taper_ratio);
  };
  // Find root_chord which fulfills - half_wing_area = half_inner_wing_area + half_outer_wing_area +
  // half_fuselage_area
  double taper_ratio_inner = newton_algorithm(taper_ratio_inner_searcher, taper_ratio_inner_max);
  if (taper_ratio_inner > taper_ratio_inner_max) {
    taper_ratio_inner = taper_ratio_inner_max;
  }
  const double taper_ratio_outer = taper_ratio / taper_ratio_inner;
  const double root_chord = root_chord_function(taper_ratio_inner);

  // Compute values for wing geometry
  const double kink_chord = root_chord * taper_ratio_inner;
  const double tip_chord = root_chord * taper_ratio;
  const double fuselage_area = 2. * eta_fuselage * half_span * root_chord;
  const double inner_wing_area = (eta_kink - eta_fuselage) * (1 + taper_ratio_inner) * half_span * root_chord;
  const double outer_wing_area = (eta_tip - eta_kink) * (taper_ratio_inner + taper_ratio) * half_span * root_chord;

  // Inner wing
  const double tan_inner_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_inner_wing_leading_edge_sweep =
      tan_inner_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);
  const double tan_inner_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan_inner_wing_quarter_chord_sweep;
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * (root_chord * taper_ratio_inner) / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);
  const double tan_outer_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord * taper_ratio_inner / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);

  // Check different conditions
  // Check sweeps leading edge and trailing edge by single equation
  // Condition 1: leading edge inner wing >= leading edge outer wing
  // Condition 2: trailing edge inner wing <= trailing edge outer wing
  // K_eta * (1-taper_ratio_2) >= taper_ratio_2 - taper_ratio -> OK
  const double K_eta = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  std::cout << "Leading edge of front wing check ... ";
  if ((K_eta * (1 - taper_ratio_inner)) >= (taper_ratio_inner - taper_ratio)) {
    std::cout << "[OK]" << std::endl;
    // Check if trailing edge is none negative
    // s*(eta_kink-eta_fuselage)*tan(sweep_leading_edge_inner) >= 1 - taper_ratio_2 -> ok
    std::cout << "Trailing edge of front wing check ... ";
    if ((half_span * (eta_kink - eta_fuselage) * tan_inner_wing_leading_edge_sweep) >= (1 - taper_ratio_inner)) {
      std::cout << "[OK]" << std::endl;
    } else {
      throwError(__FILE__,__func__,__LINE__,"Error - trailing edge condition of front wing not fulfilled (TE Sweep < 0)");
    }
  } else {
    throwError(__FILE__,__func__,__LINE__,"Error - leading edge condition of front wing not fulfilled (LE Sweep inner < LE Sweep outer).");
  }

  /* Create Sections */
  std::vector<geom2::AirfoilSection> sections;
  geom2::MultisectionSurface<geom2::AirfoilSection> wing;
  wing.is_symmetric = true;
  geom2::Vector_3 section_position;
  for (auto profile_and_thickness : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile] = profile_and_thickness;
    sections.push_back(get_rescaled_airfoil_geometry(profile_and_thickness));

    // Sections center -> fuselage
    if (eta <= eta_fuselage) {
      sections.back().origin = geom2::Point_3(0.0, 0.0, -eta * half_span);
      sections.back().set_chord_length(root_chord);
    }
    // Sections fuselage -> kink
    else if (eta > eta_fuselage && eta < eta_kink) {
      double normalized_eta = (eta - eta_fuselage) / (eta_kink - eta_fuselage);
      double x =
          std::lerp(0.0, tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span, normalized_eta);
      double y = std::lerp(0.0, tan(dihedral) * (eta_kink - eta_fuselage) * half_span, normalized_eta);
      double z = -eta * half_span;
      double section_chord = std::lerp(root_chord, taper_ratio_inner * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }
    // Sections kink -> tip
    else {
      double normalized_eta = (eta - eta_kink) / (eta_tip - eta_kink);
      double x = std::lerp(0.0, tan_outer_wing_leading_edge_sweep * (eta_tip - eta_kink) * half_span, normalized_eta) +
                 tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span;
      double y = std::lerp(0.0, tan(dihedral) * (eta_tip - eta_kink) * half_span, normalized_eta) +
                 tan(dihedral) * (eta_kink - eta_fuselage) * half_span;
      double z = -eta * half_span;
      double section_chord = std::lerp(taper_ratio_inner * root_chord, taper_ratio * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }
    sections.back().name = profile;
    wing.sections.push_back(sections.back());
    wing.normal = geom2::Direction_3(0., -1., 0.);
    // wing.origin = geom2::Point_3(0.45 * 37.5 - 0.5 * root_chord, 0., 0.);
    wing.origin = geom2::Point_3(0., 0., 0.);
  }
  // Create Sections

  return wing;
}

geom2::MultisectionSurface<geom2::AirfoilSection> Wing::calculate_kinked_back_wing_geometry(
    const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
    const double wing_area, const double span) {
  double maximum_inner_trailing_edge_sweep = get_design_maximum_inner_trailing_edge_sweep(
      config->back_relative_kink_position_calculation_mode.value(),
      config->user_back_max_inner_trailing_edge_sweep.value(),          // old converted to radians
      config->track_based_back_max_inner_trailing_edge_sweep.value());  // old converted to radians

  const double max_thickness_to_chord_ratio = get_back_max_thickness_to_chord_ratio();

  // Get normalized half span values - eta
  const double half_span = 0.5 * span;

  const double half_fuselage_width = 
    geom2::measure::width(data->fuselage, geom2::measure::length(data->fuselage) * (data->specification_longitudinal_position.value()+config->user_relative_LE_stagger.value())) * 0.5; //TODO kayra

  const double eta_fuselage = half_fuselage_width / half_span;
  double eta_centerline = 0. / half_span;
  double eta_tip = half_span / half_span;
  std::vector<double> eta_values = {eta_centerline, eta_fuselage, eta_tip};

  double eta_kink = get_design_relative_kink_position(
      config->back_relative_kink_position_calculation_mode.value(), config->user_back_relative_kink.value(),
      config->track_based_back_initial_relative_kink.value(), data->track_based_back_relative_kink.value());
  eta_values.push_back(eta_kink);

  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      get_wing_profile_and_thickness(config->back_wing_profile_and_thickness_calculation_mode.value(),
                                     config->user_back_wing_profiles_and_thickness_vector,
                                     config->back_torenbeek_jenkinson_wing_profile.value(), max_thickness_to_chord_ratio,
                                     config->back_airfoil_critical_factor.value(), data->specification_design_mach.value(),
                                     data->specification_design_altitude.value(), quarter_chord_sweep,
                                     config->back_delta_drag_divergence_to_mach_design.value(),
                                     data->sizing_point_wing_loading.value(), eta_fuselage, eta_values);

  // Geometry computation functions
  auto condition_trailing_edge_max = [=](const double inner_taper_ratio) {
    return (4. / 3. * (tan(quarter_chord_sweep) - tan(maximum_inner_trailing_edge_sweep)) * half_span *
            (eta_kink - eta_fuselage) / (1 - inner_taper_ratio));
  };

  auto root_chord_function = [=](const double inner_taper_ratio) {
    return wing_area / (2. * half_span *
                        (eta_fuselage + (1 + inner_taper_ratio) * 0.5 * (eta_kink - eta_fuselage) +
                         (inner_taper_ratio + taper_ratio) * 0.5 * (eta_tip - eta_kink)));
  };

  // Maximum possible taper ratio
  const double eta_ratio = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  const double taper_ratio_inner_max = (eta_ratio + taper_ratio) / (1 + eta_ratio);
  auto taper_ratio_inner_searcher = [=](const double inner_taper_ratio) {
    return root_chord_function(inner_taper_ratio) - condition_trailing_edge_max(inner_taper_ratio);
  };
  // Find root_chord which fulfills - half_wing_area = half_inner_wing_area + half_outer_wing_area +
  // half_fuselage_area
  double taper_ratio_inner = newton_algorithm(taper_ratio_inner_searcher, taper_ratio_inner_max);
  if (taper_ratio_inner > taper_ratio_inner_max) {
    taper_ratio_inner = taper_ratio_inner_max;
  }
  const double taper_ratio_outer = taper_ratio / taper_ratio_inner;
  const double root_chord = root_chord_function(taper_ratio_inner);

  // Compute values for wing geometry
  const double kink_chord = root_chord * taper_ratio_inner;
  const double tip_chord = root_chord * taper_ratio;
  const double fuselage_area = 2. * eta_fuselage * half_span * root_chord;
  const double inner_wing_area = (eta_kink - eta_fuselage) * (1 + taper_ratio_inner) * half_span * root_chord;
  const double outer_wing_area = (eta_tip - eta_kink) * (taper_ratio_inner + taper_ratio) * half_span * root_chord;

  // Inner wing
  const double tan_inner_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_inner_wing_leading_edge_sweep =
      tan_inner_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);
  const double tan_inner_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan_inner_wing_quarter_chord_sweep;
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * (root_chord * taper_ratio_inner) / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);
  const double tan_outer_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord * taper_ratio_inner / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);

  // Check different conditions
  // Check sweeps leading edge and trailing edge by single equation
  // Condition 1: leading edge inner wing >= leading edge outer wing
  // Condition 2: trailing edge inner wing <= trailing edge outer wing
  // K_eta * (1-taper_ratio_2) >= taper_ratio_2 - taper_ratio -> OK
  const double K_eta = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  std::cout << "Leading edge of rear wing check ... ";
  if ((K_eta * (1 - taper_ratio_inner)) >= (taper_ratio_inner - taper_ratio)) {
    std::cout << "[OK]" << std::endl;
    // Check if trailing edge is none negative
    // s*(eta_kink-eta_fuselage)*tan(sweep_leading_edge_inner) >= 1 - taper_ratio_2 -> ok
    std::cout << "Trailing edge of rear wing check ... ";
    if ((half_span * (eta_kink - eta_fuselage) * tan_inner_wing_leading_edge_sweep) >= (1 - taper_ratio_inner)) {
      std::cout << "[OK]" << std::endl;
    } else {
      throwError(__FILE__,__func__,__LINE__,"Error - trailing edge condition of rear wing not fulfilled (TE Sweep < 0)");
    }
  } else {
    throwError(__FILE__,__func__,__LINE__,"Error - leading edge condition of rear wing not fulfilled (LE Sweep inner < LE Sweep outer).");
  }

  /* Create Sections */
  std::vector<geom2::AirfoilSection> sections;
  geom2::MultisectionSurface<geom2::AirfoilSection> back_wing;
  back_wing.is_symmetric = true;
  geom2::Vector_3 section_position;
  for (auto profile_and_thickness : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile] = profile_and_thickness;
    sections.push_back(get_rescaled_airfoil_geometry(profile_and_thickness));

    // Sections center -> fuselage
    if (eta <= eta_fuselage) {
      sections.back().origin = geom2::Point_3(0.0, 0.0, -eta * half_span);
      sections.back().set_chord_length(root_chord);
    }
    // Sections fuselage -> kink
    else if (eta > eta_fuselage && eta < eta_kink) {
      double normalized_eta = (eta - eta_fuselage) / (eta_kink - eta_fuselage);
      double x =
          std::lerp(0.0, tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span, normalized_eta);
      double y = std::lerp(0.0, tan(dihedral) * (eta_kink - eta_fuselage) * half_span, normalized_eta);
      double z = -eta * half_span;
      double section_chord = std::lerp(root_chord, taper_ratio_inner * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }
    // Sections kink -> tip
    else {
      double normalized_eta = (eta - eta_kink) / (eta_tip - eta_kink);
      double x = std::lerp(0.0, tan_outer_wing_leading_edge_sweep * (eta_tip - eta_kink) * half_span, normalized_eta) +
                 tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span;
      double y = std::lerp(0.0, tan(dihedral) * (eta_tip - eta_kink) * half_span, normalized_eta) +
                 tan(dihedral) * (eta_kink - eta_fuselage) * half_span;
      double z = -eta * half_span;
      double section_chord = std::lerp(taper_ratio_inner * root_chord, taper_ratio * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }
    sections.back().name = profile;
    back_wing.sections.push_back(sections.back());
    back_wing.normal = geom2::Direction_3(0., -1., 0.);
    // back_wing.origin = geom2::Point_3(0.45 * 37.5 - 0.5 * root_chord, 0., 0.);
    back_wing.origin = geom2::Point_3(0., 0., 0.);  //TODO kayra: temporarily
  }
  // Create Sections

  return back_wing;
}


geom2::MultisectionSurface<geom2::AirfoilSection> Wing::calculate_unkinked_wing_geometry(
    const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
    const double wing_area, const double span) {
  const double max_thickness_to_chord_ratio = get_max_thickness_to_chord_ratio();

  // Get normalized half span values - eta
  const double half_span = 0.5 * span;
  const double half_fuselage_width =
      geom2::measure::width(data->fuselage, geom2::measure::length(data->fuselage) * data->specification_longitudinal_position.value()) * 0.5; //TODO kayra
  const double eta_fuselage = half_fuselage_width / half_span;
  double eta_centerline = 0. / half_span;
  double eta_tip = half_span / half_span;
  std::vector<double> eta_values = {eta_centerline, eta_fuselage, eta_tip};

  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      get_wing_profile_and_thickness(config->wing_profile_and_thickness_calculation_mode.value(),
                                     config->user_wing_profiles_and_thickness_vector,
                                     config->torenbeek_jenkinson_wing_profile.value(), max_thickness_to_chord_ratio,
                                     config->airfoil_critical_factor.value(), data->specification_design_mach.value(),
                                     data->specification_design_altitude.value(), quarter_chord_sweep,
                                     config->delta_drag_divergence_to_mach_design.value(),
                                     data->sizing_point_wing_loading.value(), eta_fuselage, eta_values);

  // Geometry computation functions

  const double root_chord = wing_area / (half_span * (2 * eta_fuselage + (1 + taper_ratio) * (eta_tip - eta_fuselage)));
  // Compute values for wing geometry

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_tip - eta_fuselage)) * (1 - taper_ratio);

  std::vector<geom2::AirfoilSection> sections;
  geom2::MultisectionSurface<geom2::AirfoilSection> wing;
  wing.is_symmetric = true;
  geom2::Vector_3 section_position;
  for (auto profile_and_thickness : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile] = profile_and_thickness;
    sections.push_back(get_rescaled_airfoil_geometry(profile_and_thickness));

    // Sections center -> fuselage
    if (eta <= eta_fuselage) {
      sections.back().origin = geom2::Point_3(0.0, 0.0, -eta * half_span);
      sections.back().set_chord_length(root_chord);
    }
    // Sections fuselage -> tip
    else {
      double normalized_eta = (eta - eta_fuselage) / (eta_tip - eta_fuselage);
      double x =
          std::lerp(0.0, tan_outer_wing_leading_edge_sweep * (eta_tip - eta_fuselage) * half_span, normalized_eta);
      double y = std::lerp(0.0, tan(dihedral) * (eta_tip - eta_fuselage) * half_span, normalized_eta);
      double z = -eta * half_span;
      double section_chord = std::lerp(root_chord, taper_ratio * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }

    sections.back().name = profile;
    wing.sections.push_back(sections.back());
    wing.normal = geom2::Direction_3(0., -1., 0.);
    wing.origin = geom2::Point_3(0., 0., 0.);
  }

  // Create Sections
  return wing;
}

geom2::MultisectionSurface<geom2::AirfoilSection> Wing::calculate_unkinked_back_wing_geometry(
    const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
    const double wing_area, const double span) {
  const double max_thickness_to_chord_ratio = get_back_max_thickness_to_chord_ratio();

  // Get normalized half span values - eta
  const double half_span = 0.5 * span;
  const double half_fuselage_width =
      geom2::measure::width(data->fuselage, geom2::measure::length(data->fuselage) * (data->specification_longitudinal_position.value()+config->user_relative_LE_stagger.value())) * 0.5; //TODO kayra
  const double eta_fuselage = half_fuselage_width / half_span;
  double eta_centerline = 0. / half_span;
  double eta_tip = half_span / half_span;
  std::vector<double> eta_values = {eta_centerline, eta_fuselage, eta_tip};

  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      get_wing_profile_and_thickness(config->back_wing_profile_and_thickness_calculation_mode.value(),
                                     config->user_back_wing_profiles_and_thickness_vector,
                                     config->back_torenbeek_jenkinson_wing_profile.value(), max_thickness_to_chord_ratio,
                                     config->back_airfoil_critical_factor.value(), data->specification_design_mach.value(),
                                     data->specification_design_altitude.value(), quarter_chord_sweep,
                                     config->back_delta_drag_divergence_to_mach_design.value(),
                                     data->sizing_point_wing_loading.value(), eta_fuselage, eta_values);

  // Geometry computation functions

  const double root_chord = wing_area / (half_span * (2 * eta_fuselage + (1 + taper_ratio) * (eta_tip - eta_fuselage)));
  // Compute values for wing geometry

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_tip - eta_fuselage)) * (1 - taper_ratio);

  std::vector<geom2::AirfoilSection> sections;
  geom2::MultisectionSurface<geom2::AirfoilSection> back_wing;
  back_wing.is_symmetric = true;
  geom2::Vector_3 section_position;
  for (auto profile_and_thickness : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile] = profile_and_thickness;
    sections.push_back(get_rescaled_airfoil_geometry(profile_and_thickness));

    // Sections center -> fuselage
    if (eta <= eta_fuselage) {
      sections.back().origin = geom2::Point_3(0.0, 0.0, -eta * half_span);
      sections.back().set_chord_length(root_chord);
    }
    // Sections fuselage -> tip
    else {
      double normalized_eta = (eta - eta_fuselage) / (eta_tip - eta_fuselage);
      double x =
          std::lerp(0.0, tan_outer_wing_leading_edge_sweep * (eta_tip - eta_fuselage) * half_span, normalized_eta);
      double y = std::lerp(0.0, tan(dihedral) * (eta_tip - eta_fuselage) * half_span, normalized_eta);
      double z = -eta * half_span;
      double section_chord = std::lerp(root_chord, taper_ratio * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }

    sections.back().name = profile;
    back_wing.sections.push_back(sections.back());
    back_wing.normal = geom2::Direction_3(0., -1., 0.);
    back_wing.origin = geom2::Point_3(0., 0., 0.);
  }

  // Create Sections
  return back_wing;
}


void Wing::flops_mass() {

  auto get_max_thickness_to_chord_ratio = [](const geom2::AirfoilSection section) {
    std::vector<double> t_to_c;
    const double chord = section.get_chord_length();
    const int steps = 50;
    const double step_width = chord / steps;
    for (size_t i = 0; i < steps; ++i) {
      t_to_c.push_back(geom2::measure::thickness(section, i * step_width) / chord);
    }
    return *std::max_element(t_to_c.begin(), t_to_c.end());
  };

  // Unit conversions and reference values for front and back wing
  const double to_foot = convertUnit(METER, FOOT, 1.0);
  const double to_meter = 1. / to_foot;
  const double taper_ratio = geom2::measure::taper_ratio(data->wing);
  const double span_ft = geom2::measure::span(data->wing) * to_foot;
  const double wing_area_sqrft = geom2::measure::reference_area(data->wing) * to_foot * to_foot;
  const double aspect_ratio = geom2::measure::aspect_ratio(data->wing);
  const double t_to_c_root = get_max_thickness_to_chord_ratio(data->wing.sections.front());
  const double t_to_c_tip = get_max_thickness_to_chord_ratio(data->wing.sections.back());
  const double quarter_chord_sweep =
      geom2::measure::sweep(data->wing, 0.25 * span_ft * to_meter, 0.25);  // measure at 50% half span -> 25% of span

  const double to_lb = convertUnit(KILO, GRAM, NOPREFIX, POUND, 1.0);
  const double to_kg = 1 / to_lb;

  const double back_taper_ratio = geom2::measure::taper_ratio(data->back_wing);
  const double back_span_ft = geom2::measure::span(data->back_wing) * to_foot;
  const double back_wing_area_sqrft = geom2::measure::reference_area(data->back_wing) * to_foot * to_foot;
  const double back_aspect_ratio = geom2::measure::aspect_ratio(data->back_wing);
  const double back_t_to_c_root = get_max_thickness_to_chord_ratio(data->back_wing.sections.front());
  const double back_t_to_c_tip = get_max_thickness_to_chord_ratio(data->back_wing.sections.back());
  const double back_quarter_chord_sweep =
      geom2::measure::sweep(data->back_wing, 0.25 * back_span_ft * to_meter, 0.25);  // measure at 50% half span -> 25% of span

  
  // It is assumed for preliminary design stage that each wing carries load proportinal to its wing area  //TODO kayra
      
  // PCTL for front wing
  const double PCTL = wing_area_sqrft / (wing_area_sqrft + back_wing_area_sqrft); //TODO kayra:degismeli

  // PCTL for back wing  
  const double back_PCTL = back_wing_area_sqrft / (wing_area_sqrft + back_wing_area_sqrft);   //TODO kayra: degismeli

  const double dg = data->mtom_mass_properties.data["mass"].value() * to_lb;  //* PCTL; //TODO kayra: bu isi de sor

  const double back_dg = data->mtom_mass_properties.data["mass"].value() * to_lb; //* back_PCTL; //TODO kayra: bu isi de sor



  myRuntimeInfo->out << "Front Wing mass estimation using ... FLOPS" << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;

  myRuntimeInfo->warn << "Following values are Hard coded!" << std::endl;
  myRuntimeInfo->warn << "================================" << std::endl;
  myRuntimeInfo->warn << "\twing_mounted_engines" << std::endl;
  myRuntimeInfo->warn << "================================" << std::endl;




  // Tangent of 3/4 chord wing sweep of front wing
  double TLAM = tan(quarter_chord_sweep) - 2.0 * (1.0 - taper_ratio) / (aspect_ratio * (1.0 + taper_ratio));
  // Sine of 3/4 chord wing sweep angle of front wing
  double SLAM = TLAM / (sqrt(1 + TLAM * TLAM));

  // Factor Eq. 15 of front wing
  double C4 = 1.0 - 0.5 * config->flops_faert.value();

  // Factor Eq. 16 of front wing
  double C6 = 0.5 * config->flops_faert.value() - 0.16 * config->flops_fstrt.value();

  // Aspect ratio defined factor Eq. 17 of front wing
  double CAYA = aspect_ratio > 5. ? aspect_ratio - 5.0 : 0.;
  // Wing sweep and aeroelastic factor of front wing
  double CAYL = (1.0 - SLAM * SLAM) * (1.0 + C6 * SLAM * SLAM + 0.03 * CAYA * C4 * SLAM);

  // Wing strut bracing factor of front wing (for cantilever w/o wing strut)
  double EMS = 1.0 - 0.25 * config->flops_fstrt.value();
  // Weighted average of the wing thickness to chord ratio of front wing (Jenkinson 1999 formulation)
  double TCA = 0.25 * (3. * t_to_c_tip + t_to_c_root);
  // wing equivalent bending material factor of front wing Dimension [-]
  double BT = 0.215 * (0.37 + 0.7 * taper_ratio) * pow(span_ft * span_ft / wing_area_sqrft, EMS) / (CAYL * TCA);

  // Total wing bending material weight of front wing

  // A1 - Table 1 Flops wing weight constants for transport and hwb
  const double A1 = 8.80;

  // A2
  const double A2 = 6.25;

  // ULF - Structural Ultimate load factor - default 3.75
  const double ULF = 3.75;

  // CAYF - only single fuselage usage
  const double CAYF = 1.0;

  // VARSWP - no variable sweep allowed
  const double VARSWP = 0.0;
  // VFACT
  const double VFACT = 1.0;


  // Wing bending material weight of front wing, not including the effects of inertia [lb]
  const double W1NIR = A1 * BT * (1.0 + sqrt(A2 / span_ft)) * ULF * span_ft *
                       (1.0 - 0.4 * config->flops_fcomp.value()) * (1.0 - 0.1 * config->flops_faert.value()) * CAYF *
                       VFACT * PCTL / 1000000.0;

  // Total wing shear material and control surface weight of front wing
  const double A3 = 0.68;
  const double A4 = 0.34;
  const double A5 = 0.6;

  // total movable wing surface area of front wing including flaps, elevators, spoilers etc [ft^2]

  const double SFLAP = get_control_device_area_in_fraction_of_wing_area() *
                       wing_area_sqrft;  // set to 20% of wing_area - testing only TODO

  const double W2 = A3 * (1.0 - 0.17 * config->flops_fcomp.value()) * pow(SFLAP, A4) * pow(dg, A5);

  // Total wing miscellaneous items weight of front wing
  const double A6 = 0.035;
  const double A7 = 1.50;

  const double W3 = A6 * (1.0 - 0.3 * config->flops_fcomp.value()) * pow(wing_area_sqrft, A7);

  // Wing bending material weight intertia relief adjustment of front wing
  const int NEW = data->specification_number_of_wing_mounted_engines;
  const double CAYE = 1.0 - 0.03 * NEW;

  const double W1 = (dg * CAYE * W1NIR + W2 + W3) / (1.0 + W1NIR) - W2 - W3;

  // Total aft body weight for hybrid wing body aircraft - here no HWB
  const double W4 = 0.0;

  // Total wing weight of front wing
  const double wing_mass = config->flops_technology_factor.value() * (W1 + W2 + W3 + W4) * to_kg;

  myRuntimeInfo->out << "Calculated front wing mass ... " << wing_mass << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing bending material mass of front wing... " << config->flops_technology_factor.value() * W1 * to_kg << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing shear material and control surface mass of front wing ... " << config->flops_technology_factor.value() * W2 * to_kg << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing miscellaneous items mass of front wing ... " << config->flops_technology_factor.value() * W3 * to_kg << "[kg]" << std::endl;



  myRuntimeInfo->out << "Back Wing mass estimation using ... FLOPS" << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;

  myRuntimeInfo->warn << "Following values are Hard coded!" << std::endl;
  myRuntimeInfo->warn << "================================" << std::endl;
  myRuntimeInfo->warn << "\twing_mounted_engines" << std::endl;
  myRuntimeInfo->warn << "================================" << std::endl;


  // Tangent of 3/4 chord wing sweep of back wing
  double back_TLAM = tan(back_quarter_chord_sweep) - 2.0 * (1.0 - back_taper_ratio) / (back_aspect_ratio * (1.0 + back_taper_ratio));
  // Sine of 3/4 chord wing sweep angle of back wing
  double back_SLAM = back_TLAM / (sqrt(1 + back_TLAM * back_TLAM));

  // Factor Eq. 15 of back wing
  double back_C4 = 1.0 - 0.5 * config->back_flops_faert.value();

  // Factor Eq. 16 of back wing
  double back_C6 = 0.5 * config->back_flops_faert.value() - 0.16 * config->back_flops_fstrt.value();

  // Aspect ratio defined factor Eq. 17 of back wing
  double back_CAYA = back_aspect_ratio > 5. ? back_aspect_ratio - 5.0 : 0.;
  // Wing sweep and aeroelastic factor of back wing
  double back_CAYL = (1.0 - back_SLAM * back_SLAM) * (1.0 + back_C6 * back_SLAM * back_SLAM + 0.03 * back_CAYA * back_C4 * back_SLAM);

  // Wing strut bracing factor of back wing(for cantilever w/o wing strut)
  double back_EMS = 1.0 - 0.25 * config->back_flops_fstrt.value();
  // Weighted average of the wing thickness to chord ratio of back wing(Jenkinson 1999 formulation)
  double back_TCA = 0.25 * (3. * back_t_to_c_tip + back_t_to_c_root);
  // wing equivalent bending material factor of back wing Dimension [-]
  double back_BT = 0.215 * (0.37 + 0.7 * back_taper_ratio) * pow(back_span_ft * back_span_ft / back_wing_area_sqrft, back_EMS) / (back_CAYL * back_TCA);

  // Total wing bending material weight of back wing
  //TODO kayra: I keep these variables with different names altough they are same with front wing values, in order to keep maintenance easier if some of them should change in the future.
  // A1 - Table 1 Flops wing weight constants for transport and hwb
  const double back_A1 = 8.80;

  // A2
  const double back_A2 = 6.25;

  // ULF - Structural Ultimate load factor - default 3.75
  const double back_ULF = 3.75;

  // CAYF - only single fuselage usage
  const double back_CAYF = 1.0;

  // VARSWP - no variable sweep allowed
  const double back_VARSWP = 0.0;
  // VFACT
  const double back_VFACT = 1.0;


  // Wing bending material weight of back wing, not including the effects of inertia [lb]
  const double back_W1NIR = back_A1 * back_BT * (1.0 + sqrt(back_A2 / back_span_ft)) * back_ULF * back_span_ft *
                       (1.0 - 0.4 * config->back_flops_fcomp.value()) * (1.0 - 0.1 * config->back_flops_faert.value()) * back_CAYF *
                       back_VFACT * back_PCTL / 1000000.0;

  // Total wing shear material and control surface weight of back wing
  const double back_A3 = 0.68;
  const double back_A4 = 0.34;
  const double back_A5 = 0.6;

  // total movable wing surface area of back wing including flaps, elevators, spoilers etc [ft^2]

  const double back_SFLAP = get_back_control_device_area_in_fraction_of_back_wing_area() *
                       back_wing_area_sqrft;  // set to 20% of wing_area - testing only TODO

  const double back_W2 = back_A3 * (1.0 - 0.17 * config->back_flops_fcomp.value()) * pow(back_SFLAP, back_A4) * pow(back_dg, back_A5);

  // Total wing miscellaneous items weight of back wing
  const double back_A6 = 0.035;
  const double back_A7 = 1.50;

  const double back_W3 = back_A6 * (1.0 - 0.3 * config->back_flops_fcomp.value()) * pow(back_wing_area_sqrft, back_A7);

  // Wing bending material weight intertia relief adjustment of back wing
  const int back_NEW = data->specification_number_of_wing_mounted_engines;
  const double back_CAYE = 1.0 - 0.03 * back_NEW;

  const double back_W1 = (back_dg * back_CAYE * back_W1NIR + back_W2 + back_W3) / (1.0 + back_W1NIR) - back_W2 - back_W3;

  // Total aft body weight for hybrid wing body aircraft - here no HWB
  const double back_W4 = 0.0;

  // Total wing weight of back wing
  const double back_wing_mass = config->back_flops_technology_factor.value() * (back_W1 + back_W2 + back_W3 + back_W4) * to_kg;

  myRuntimeInfo->out << "Calculated back wing mass ... " << back_wing_mass << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing bending material mass of back wing ... " << config->back_flops_technology_factor.value() * back_W1 * to_kg << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing shear material and control surface mass of back wing ... " << config->back_flops_technology_factor.value() * back_W2 * to_kg << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing miscellaneous items mass of back wing ... " << config->back_flops_technology_factor.value() * back_W3 * to_kg << "[kg]" << std::endl;

  myRuntimeInfo->out << "Overall Tandem Wing Set mass estimation by adding both wings masses calculated with FLOPS method ..." << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;

  const double tandem_wing_set_mass = wing_mass + back_wing_mass;

  myRuntimeInfo->out << "Calculated overall tandem wing set mass ... " << tandem_wing_set_mass << "[kg]" << std::endl;

  data->wing_mass_properties.data["mass"] = tandem_wing_set_mass;
  //data->back_wing_mass_properties.data["mass"] = back_wing_mass;  //TODO kayra: olmamasi lazim



  // Calculate wing center of gravity of front wing according to Howe 2005 page 252
  myRuntimeInfo->out << "Calculate Center of Gravity of front wing ..." << std::endl;
  const double half_span_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  const double mean_chord = geom2::measure::reference_area(data->wing) / geom2::measure::span(data->wing);
  const double mac = geom2::measure::chord(data->wing, half_span_mac_position);
  const geom2::Point_3 leading_edge_point = geom2::measure::offset_LE(data->wing, half_span_mac_position); //TODO kayra: if twist this should be adjusted
  const double CoG_x = data->wing_position.x.value() + leading_edge_point.x() + 0.25 * mac +
                                                 0.1 * mean_chord;
  // Transform values into global coordinate system z-> -y z
  const double CoG_y = data->wing_position.y.value() + fabs(leading_edge_point.z()) -
                                                 fabs(leading_edge_point.z());
  const double CoG_z = data->wing_position.z.value() + leading_edge_point.y();

  myRuntimeInfo->out << "CG Position of front wing..." << std::endl;
  myRuntimeInfo->out << "x ... " << CoG_x << std::endl;
  myRuntimeInfo->out << "y ... " << CoG_y << std::endl;
  myRuntimeInfo->out << "z ... " << CoG_z << std::endl;




  // Calculate wing center of gravity of back wing according to Howe 2005 page 252
  myRuntimeInfo->out << "Calculate Center of Gravity of back wing..." << std::endl;
  const double back_half_span_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->back_wing);
  const double back_mean_chord = geom2::measure::reference_area(data->back_wing) / geom2::measure::span(data->back_wing);
  const double back_mac = geom2::measure::chord(data->back_wing, back_half_span_mac_position);
  const geom2::Point_3 back_leading_edge_point = geom2::measure::offset_LE(data->back_wing, back_half_span_mac_position); //TODO kayra: if twist this should be adjusted
  const double back_CoG_x = data->wing_position.x.value() + data->back_wing.origin.x() + back_leading_edge_point.x() + 0.25 * back_mac +
                                                 0.1 * back_mean_chord;
  // Transform values into global coordinate system z-> -y z
  const double back_CoG_y = data->wing_position.y.value() + data->back_wing.origin.y() + fabs(back_leading_edge_point.z()) -
                                                 fabs(back_leading_edge_point.z());
  const double back_CoG_z = data->wing_position.z.value() + data->back_wing.origin.z() + back_leading_edge_point.y();

  myRuntimeInfo->out << "CG Position of back wing..." << std::endl;
  myRuntimeInfo->out << "x ... " << back_CoG_x << std::endl;
  myRuntimeInfo->out << "y ... " << back_CoG_y << std::endl;
  myRuntimeInfo->out << "z ... " << back_CoG_z << std::endl;

 
  // Calculate center of gravity of tandem wing set using mass weighted CoG formula //TODO kayra
  myRuntimeInfo->out << "Calculate Center of Gravity of overall tandem wing set ..." << std::endl;
  const double wingset_CoG_x = (CoG_x * wing_mass + back_CoG_x * back_wing_mass) / tandem_wing_set_mass;
  const double wingset_CoG_y = (CoG_y * wing_mass + back_CoG_y * back_wing_mass) / tandem_wing_set_mass;
  const double wingset_CoG_z = (CoG_z * wing_mass + back_CoG_z * back_wing_mass) / tandem_wing_set_mass;

  
  data->wing_mass_properties.data["x"].set_value(wingset_CoG_x);
  data->wing_mass_properties.data["y"].set_value(wingset_CoG_y);
  data->wing_mass_properties.data["z"].set_value(wingset_CoG_z);

  myRuntimeInfo->out << "CG Position of Tandem Wing Set ..." << std::endl;
  myRuntimeInfo->out << "x ... " << data->wing_mass_properties.data["x"].value() << std::endl;
  myRuntimeInfo->out << "y ... " << data->wing_mass_properties.data["y"].value() << std::endl;
  myRuntimeInfo->out << "z ... " << data->wing_mass_properties.data["z"].value() << std::endl;


  //TODO kayra: yedek olarak biraktim
  /*data->wing_mass_properties.data["x"].set_value(data->wing_position.x.value() + leading_edge_point.x() + 0.25 * mac +
                                                 0.1 * mean_chord);
  // Transform values into global coordinate system z-> -y z
  data->wing_mass_properties.data["y"].set_value(data->wing_position.y.value() + fabs(leading_edge_point.z()) -
                                                 fabs(leading_edge_point.z()));
  data->wing_mass_properties.data["z"].set_value(data->wing_position.z.value() + leading_edge_point.y());

  myRuntimeInfo->out << "CG Position of front wing..." << std::endl;
  myRuntimeInfo->out << "x ... " << data->wing_mass_properties.data["x"].value() << std::endl;
  myRuntimeInfo->out << "y ... " << data->wing_mass_properties.data["y"].value() << std::endl;
  myRuntimeInfo->out << "z ... " << data->wing_mass_properties.data["z"].value() << std::endl;


  data->back_wing_mass_properties.data["x"].set_value(data->back_wing.origin.x() + back_leading_edge_point.x() + 0.25 * back_mac +
                                                 0.1 * back_mean_chord);
  // Transform values into global coordinate system z-> -y z
  data->back_wing_mass_properties.data["y"].set_value(data->back_wing.origin.y() + fabs(leading_edge_point.z()) -
                                                 fabs(leading_edge_point.z()));
  data->back_wing_mass_properties.data["z"].set_value(data->back_wing.origin.z() + leading_edge_point.y());

  myRuntimeInfo->out << "CG Position of back wing..." << std::endl;
  myRuntimeInfo->out << "x ... " << data->back_wing_mass_properties.data["x"].value() << std::endl;
  myRuntimeInfo->out << "y ... " << data->back_wing_mass_properties.data["y"].value() << std::endl;
  myRuntimeInfo->out << "z ... " << data->back_wing_mass_properties.data["z"].value() << std::endl;*/

}


/**
 * \brief Chiozzotto weight estimation relationship mass estimation - assumes symmetric engine distribution on wings
 *
 */
// TODO kayra: bunda asla kink olmazsa sonuc cikmiyor sikintili bir yöntem
void Wing::chiozzotto_wer_mass() {
  myRuntimeInfo->out << "Front Wing mass estimation using ... CHIOZZOTTO" << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;


  std::map<std::string, std::tuple<double, double>> min_max_mtom = {
      {"AL ASW", {20000, 250000}},   {"AL USW", {20000, 250000}},   {"AL FSW", {20000, 250000}},
      {"CFRP ASW", {20000, 250000}}, {"CFRP USW", {20000, 250000}}, {"CFRP FSW", {20000, 250000}}};
  std::map<std::string, std::tuple<double, double>> min_max_w_to_s = {
      {"AL ASW", {3000, 8000}},   {"AL USW", {3000, 8000}},   {"AL FSW", {3000, 8000}},
      {"CFRP ASW", {3000, 8000}}, {"CFRP USW", {3000, 8000}}, {"CFRP FSW", {3000, 8000}}};
  std::map<std::string, std::tuple<double, double>> min_max_aspect_ratio = {
      {"AL ASW", {10, 20}},   {"AL USW", {10, 20}},   {"AL FSW", {8, 16}},
      {"CFRP ASW", {10, 20}}, {"CFRP USW", {10, 20}}, {"CFRP FSW", {8, 16}}};
  std::map<std::string, std::tuple<double, double>> min_max_sweep = {{"AL ASW", {10, 40}},  {"AL USW", {0, 10}},
                                                                     {"AL FSW", {-5, -25}}, {"CFRP ASW", {10, 40}},
                                                                     {"CFRP USW", {0, 10}}, {"CFRP FSW", {-5, -25}}};
  std::map<std::string, std::tuple<double, double>> min_max_t_to_c = {
      {"AL ASW", {0.08, 0.14}},   {"AL USW", {0.08, 0.18}},   {"AL FSW", {0.08, 0.16}},
      {"CFRP ASW", {0.08, 0.14}}, {"CFRP USW", {0.08, 0.18}}, {"CFRP FSW", {0.08, 0.16}}};
  std::map<std::string, std::tuple<double, double>> min_max_v_eas = {
      {"AL ASW", {130, 200}},   {"AL USW", {130, 200}},   {"AL FSW", {130, 200}},
      {"CFRP ASW", {130, 200}}, {"CFRP USW", {130, 200}}, {"CFRP FSW", {130, 200}}};
  std::map<std::string, std::tuple<double, double>> min_max_taper_ratio = {
      {"AL ASW", {0.1, 0.5}},   {"AL USW", {0.1, 0.5}},   {"AL FSW", {0.1, 0.5}},
      {"CFRP ASW", {0.1, 0.5}}, {"CFRP USW", {0.1, 0.5}}, {"CFRP FSW", {0.1, 0.5}}};

  /**< Parameter containing a table of the constants and exponents of conventional Aircraft Covers*/
  std::map<std::string, std::tuple<double, double, double, double, double, double, double, double>>
      lookup_table_covers = {{"AL ASW", {2.14e-3, 1.345, -0.623, 1.286, -1.601, -0.830, -0.010, 0.104}},
                             {"AL USW", {2.66e-3, 1.320, -0.731, 1.370, 6.733, -0.945, 0.110, 0.305}},
                             {"AL FSW", {2.82e-5, 1.373, -1.152, 2.006, -5.507, -1.549, 1.036, 0.698}},
                             {"CFRP ASW", {9.47e-4, 1.440, -0.699, 1.233, -1.044, -0.808, 0.017, 0.518}},
                             {"CFRP USW", {1.29e-3, 1.396, -0.929, 1.4481, 0.667, -1.051, 0.266, 0.633}},
                             {"CFRP FSW", {3.92e-5, 1.401, -1.111, 1.874, -4.084, -1.436, 0.889, 0.751}}};

  /**< Parameter containing a table of constants and exponents of conventional Aircraft spars and ribs*/
  std::map<std::string, std::tuple<double, double, double, double, double, double, double, double>>
      lookup_table_spars_and_ribs = {{"AL ASW", {1.86, 1.384, -0.972, -0.072, -0.072, 0.486, 0.043, -0.126}},
                                     {"AL USW", {2.99, 1.376, -1.090, 0.008, 5.376, 0.423, 0.112, -0.070}},
                                     {"AL FSW", {9.82e-2, 1.362, -1.319, 0.598, -3.711, -0.126, 0.673, 0.391}},
                                     {"CFRP ASW", {1.50e-1, 1.411, -0.832, 0.105, -0.235, 0.301, 0.057, 0.101}},
                                     {"CFRP USW", {2.85e-1, 1.391, -0.994, 0.210, 5.854, 0.194, 0.164, 0.189}},
                                     {"CFRP FSW", {2.31e-2, 1.370, -1.174, 0.650, -2.750, -0.204, 0.620, 0.454}}};

  /**< Parameter containing a table of constants  and exponents of conventional Aircraft engine relief*/
  std::map<std::string, std::tuple<double, double, double>> lookup_table_engine_relief = {
      {"AL ASW", {8.41, 3.017, 0.817}},   {"AL USW", {12.5, 2.177, 1.011}},   {"AL FSW", {25.4, 1.941, 1.204}},
      {"CFRP ASW", {15.3, 3.104, 0.909}}, {"CFRP USW", {13.6, 2.569, 0.910}}, {"CFRP FSW", {24.7, 2.117, 1.148}}};


  const double ref_area = geom2::measure::reference_area(data->wing);             //TODO kayra: düzenleme
  const double back_ref_area = geom2::measure::reference_area(data->back_wing);

  const double area_portion = ref_area / (ref_area + back_ref_area);
  const double back_area_portion = back_ref_area / (ref_area + back_ref_area);

  // Chiozzotto Engine relief factor k_e (5.5 - Page 116)
  // IF NUMBER OF ENGINES ON WING > 2 ->
  // double engine_relative_spanwise_position = 0.3;  // BOOKMARK
  // Vector with data for one half ->
  // More than one engine -> k_e = k_e_1 * k_e_2

  double wing_loading = data->sizing_point_wing_loading.value();  //TODO kayra: düzenleme yok zaten kendini buluyor
  if (data->sizing_point_wing_loading.unit() == "kg/m^2") {
    wing_loading *= G_FORCE;
  }

  atmosphere isa;
  const double mtom = data->mtom_mass_properties.data["mass"].value() * area_portion;  //TODO kayra: düzenleme  : lift generation prop. to its area portion: first assumption
  const double aspect_ratio = geom2::measure::aspect_ratio(data->wing);
  const double taper_ratio = geom2::measure::taper_ratio(data->wing);
  const double half_span = geom2::measure::span(data->wing) * 0.5;
  const double speed_of_sound_at_zero = isa.getSpeedOfSoundISA(0.0);
  const double pressure_at_zero = isa.getPressureISA(0.0);
  const double pressure_at_design_altitude = isa.getPressureISA(data->specification_design_altitude.value());

  std::vector<std::tuple<double, double>> engines_mass_and_eta = data->wing_mounted_engine_data;

  // Calculation of EAS Design
  const double v_eas_design = data->specification_design_mach.value() * speed_of_sound_at_zero *
                              sqrt(pressure_at_design_altitude / pressure_at_zero);

  // No LRA sweep existent - use 1/4 chord sweep (deviation from chiozzotto P.112 -> referred to figure 5.3)
  const double quarter_chord_sweep = -geom2::measure::sweep(data->wing, 0.5 * half_span, 0.25);

  // Get material selection
  std::string material = config->chiozzotto_material.value() + " ";
  if (quarter_chord_sweep > ACCURACY_LOW) {
    material += "ASW";
  } else if (quarter_chord_sweep < -ACCURACY_LOW) {
    material += "FSW";
  } else {
    material += "USW";
  }
  myRuntimeInfo->out << "Wing material used for front wing ... " << material << std::endl;

  // Get design relative kink -
  double t_to_c_kink = 0.0;   //TODO kayra : düzeltilecek
  double eta_kink = 0.35;
  if (is_wing_kinked_check()) {
    eta_kink = get_design_relative_kink_position(
        config->relative_kink_position_calculation_mode.value(), config->user_relative_kink.value(),
        config->track_based_initial_relative_kink.value(), data->track_based_relative_kink.value());
  }

  // Find thickness to chord ratio
  const double chord_at_kink = geom2::measure::chord(data->wing, eta_kink * half_span);

  double eta_position_last = 0.;
  double max_thickness_last = 0.0;
  double eta_position_next = 0.;
  double max_thickness_next = 0.0;
  for (size_t idx = 0; idx < data->wing.sections.size() - 1; ++idx) {
    eta_position_last = fabs(data->wing.sections.at(idx).origin.z()) / half_span;
    eta_position_next = fabs(data->wing.sections.at(idx + 1).origin.z()) / half_span;
    // If kink at a specified section
    if (fabs(eta_position_last - eta_kink) < ACCURACY_MEDIUM) {
      t_to_c_kink = geom2::measure::thickness_max(data->wing.sections.at(idx)) / chord_at_kink;
    } else if (fabs(eta_position_next - eta_kink) < ACCURACY_MEDIUM) {
      t_to_c_kink = geom2::measure::thickness_max(data->wing.sections.at(idx)) / chord_at_kink;
    } else if (eta_position_last > eta_kink && eta_position_next < eta_kink) {
      double max_thickness_last = geom2::measure::thickness_max(data->wing.sections.at(idx));
      double max_thickness_next = geom2::measure::thickness_max(data->wing.sections.at(idx));
      double dimensionless_position = (eta_kink - eta_position_last) / (eta_position_last - eta_position_next);
      t_to_c_kink = std::lerp(max_thickness_last, max_thickness_next, dimensionless_position) / chord_at_kink;
    } else {
      // Do nothing
    }
  }

  // Check inputs
  if (!(chiozzotto_wer_check_limits(min_max_mtom.at(material), mtom))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... mtom front wing portion ... " << mtom << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_w_to_s.at(material), wing_loading))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... front wing_loading ... " << wing_loading << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_aspect_ratio.at(material), aspect_ratio))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... front wings aspect ratio ... " << aspect_ratio << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_taper_ratio.at(material), taper_ratio))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... front wings taper ratio ... " << taper_ratio << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_sweep.at(material), quarter_chord_sweep * geom2::detail::to_degrees))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... front wings sweep ... "
                        << quarter_chord_sweep * geom2::detail::to_degrees << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_t_to_c.at(material), t_to_c_kink))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... front wings thickness to chord ... " << t_to_c_kink
                        << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_v_eas.at(material), v_eas_design))) {
    myRuntimeInfo->warn << "Chiozzotto limit check for front wing out of bounds ... equivalent airspeed at design point ... "
                        << v_eas_design << std::endl;
  }

  const auto [C_er, E_eta_er, E_K_er] = lookup_table_engine_relief[material];

  // Engine relief factor
  double k_e = 1.;
  for (auto engine : engines_mass_and_eta) {
    const auto [engine_mass, engine_spanwise_position] = engine;
    /* If engines exist, the spanwise position is absolute, if engines not exist, the default values are in rel. half span */
    double engine_relative_spanwise_position = data->engines_exist ? std::fabs(engine_spanwise_position)/half_span : engine_spanwise_position;
    k_e *= (1 - C_er * pow(engine_relative_spanwise_position, E_eta_er) * pow(engine_mass / mtom, E_K_er));
  }
  // Secondary structure mass
  double mass_secondary_structure = 0.0443 * mtom * config->chiozzotto_technology_factor.value();

  // Mass of covers
  const auto [C_cov, mtom_cov, w_to_s_cov, ar_cov, cos_lambda_cov, t_to_c_kink_cov, v_eas_design_cov,
              one_plus_taper_cov] = lookup_table_covers[material];
  double mass_covers = (k_e * C_cov * pow(mtom, mtom_cov) * pow(wing_loading, w_to_s_cov) * pow(aspect_ratio, ar_cov) *
                        pow(cos(quarter_chord_sweep), cos_lambda_cov) * pow(t_to_c_kink, t_to_c_kink_cov) *
                        pow(v_eas_design, v_eas_design_cov) * pow(1 + taper_ratio, one_plus_taper_cov)) *
                       config->chiozzotto_technology_factor.value();

  // Mass of spars and ribs
  const auto [C_sr, mtom_sr, w_to_s_sr, ar_sr, cos_lambda_sr, t_to_c_kink_sr, v_eas_design_sr, one_plus_taper_sr] =
      lookup_table_spars_and_ribs[material];
  double mass_spars_and_ribs = (C_sr * pow(mtom, mtom_sr) * pow(wing_loading, w_to_s_sr) * pow(aspect_ratio, ar_sr) *
                                pow(cos(quarter_chord_sweep), cos_lambda_sr) * pow(t_to_c_kink, t_to_c_kink_sr) *
                                pow(v_eas_design, v_eas_design_sr) * pow(1 + taper_ratio, one_plus_taper_sr)) *
                               config->chiozzotto_technology_factor.value();

  double wing_mass = mass_covers + mass_spars_and_ribs + mass_secondary_structure;

  myRuntimeInfo->out << "Calculated front wing mass ... " << wing_mass << "[kg]" << std::endl;
  myRuntimeInfo->out << "Front Wing mass spars and ribs ..." << mass_spars_and_ribs << "[kg]" << std::endl;
  myRuntimeInfo->out << "Front Wing mass covers  ..." << mass_covers << "[kg]" << std::endl;
  myRuntimeInfo->out << "Front Wing mass secondary structure ..." << mass_secondary_structure << "[kg]" << std::endl;


  myRuntimeInfo->out << "Back Wing mass estimation using ... CHIOZZOTTO" << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;
  
  
  // Chiozzotto Engine relief factor k_e (5.5 - Page 116)
  // IF NUMBER OF ENGINES ON WING > 2 ->
  // double engine_relative_spanwise_position = 0.3;  // BOOKMARK
  // Vector with data for one half ->
  // More than one engine -> k_e = k_e_1 * k_e_2

  double back_wing_loading = data->sizing_point_wing_loading.value();    //TODO kayra: düzenleme
  if (data->sizing_point_wing_loading.unit() == "kg/m^2") {
    back_wing_loading *= G_FORCE;
  }

  const double back_mtom = data->mtom_mass_properties.data["mass"].value() * back_area_portion;    //TODO kayra: düzenleme
  const double back_aspect_ratio = geom2::measure::aspect_ratio(data->back_wing);
  const double back_taper_ratio = geom2::measure::taper_ratio(data->back_wing);
  const double back_half_span = geom2::measure::span(data->back_wing) * 0.5;

  std::vector<std::tuple<double, double>> back_engines_mass_and_eta = data->wing_mounted_engine_data;

  // Calculation of EAS Design same with front wing

  // No LRA sweep existent - use 1/4 chord sweep (deviation from chiozzotto P.112 -> referred to figure 5.3)
  const double back_quarter_chord_sweep = -geom2::measure::sweep(data->back_wing, 0.5 * back_half_span, 0.25);

  // Get material selection
  std::string back_material = config->back_chiozzotto_material.value() + " ";
  if (back_quarter_chord_sweep > ACCURACY_LOW) {
    back_material += "ASW";
  } else if (back_quarter_chord_sweep < -ACCURACY_LOW) {
    back_material += "FSW";
  } else {
    back_material += "USW";
  }
  myRuntimeInfo->out << "Wing material used for back wing ... " << back_material << std::endl;

  // Get design relative kink -
  double back_t_to_c_kink = 0.0;
  double back_eta_kink = 0.35;
  if (is_back_wing_kinked_check()) {
    back_eta_kink = get_design_relative_kink_position(
        config->back_relative_kink_position_calculation_mode.value(), config->user_back_relative_kink.value(),
        config->track_based_back_initial_relative_kink.value(), data->track_based_back_relative_kink.value());
  }

  // Find thickness to chord ratio
  const double back_chord_at_kink = geom2::measure::chord(data->back_wing, back_eta_kink * back_half_span);

  double back_eta_position_last = 0.;
  double back_max_thickness_last = 0.0;
  double back_eta_position_next = 0.;
  double back_max_thickness_next = 0.0;
  for (size_t idx = 0; idx < data->back_wing.sections.size() - 1; ++idx) {
    back_eta_position_last = fabs(data->back_wing.sections.at(idx).origin.z()) / back_half_span;
    back_eta_position_next = fabs(data->back_wing.sections.at(idx + 1).origin.z()) / back_half_span;
    // If kink at a specified section
    if (fabs(back_eta_position_last - back_eta_kink) < ACCURACY_MEDIUM) {
      back_t_to_c_kink = geom2::measure::thickness_max(data->back_wing.sections.at(idx)) / back_chord_at_kink;
    } else if (fabs(back_eta_position_next - back_eta_kink) < ACCURACY_MEDIUM) {
      back_t_to_c_kink = geom2::measure::thickness_max(data->back_wing.sections.at(idx)) / back_chord_at_kink;
    } else if (back_eta_position_last > back_eta_kink && back_eta_position_next < back_eta_kink) {
      double back_max_thickness_last = geom2::measure::thickness_max(data->back_wing.sections.at(idx));
      double back_max_thickness_next = geom2::measure::thickness_max(data->back_wing.sections.at(idx));
      double back_dimensionless_position = (back_eta_kink - back_eta_position_last) / (back_eta_position_last - back_eta_position_next);
      back_t_to_c_kink = std::lerp(back_max_thickness_last, back_max_thickness_next, back_dimensionless_position) / back_chord_at_kink;
    } else {
      // Do nothing
    }
  }

  // Check inputs       //TODO kayra: bundan assagida bir seyi ikiyle carpip sonuc alabilirsin sanirim iki wing massi ile
  if (!(chiozzotto_wer_check_limits(min_max_mtom.at(back_material), back_mtom))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... mtom ... " << back_mtom << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_w_to_s.at(back_material), back_wing_loading))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... wing_loading ... " << back_wing_loading << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_aspect_ratio.at(back_material), back_aspect_ratio))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... aspect ratio ... " << back_aspect_ratio << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_taper_ratio.at(back_material), back_taper_ratio))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... taper ratio ... " << back_taper_ratio << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_sweep.at(back_material), back_quarter_chord_sweep * geom2::detail::to_degrees))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... sweep ... "
                        << back_quarter_chord_sweep * geom2::detail::to_degrees << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_t_to_c.at(back_material), back_t_to_c_kink))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... thickness to chord ... " << back_t_to_c_kink
                        << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_v_eas.at(back_material), v_eas_design))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... equivalent airspeed at design point ... "
                        << v_eas_design << std::endl;
  }

  const auto [back_C_er, back_E_eta_er, back_E_K_er] = lookup_table_engine_relief[back_material];
 
  
  // Engine relief factor
  double back_k_e = 1.;
  for (auto engine : back_engines_mass_and_eta) {
    const auto [back_engine_mass, back_engine_spanwise_position] = engine;
    /* If engines exist, the spanwise position is absolute, if engines not exist, the default values are in rel. half span */
    double back_engine_relative_spanwise_position = data->engines_exist ? std::fabs(back_engine_spanwise_position)/back_half_span : back_engine_spanwise_position;
    //TODO kayra: engine existin arka kanat icin farkli olmasi gerekebilir
    back_k_e *= (1 - back_C_er * pow(back_engine_relative_spanwise_position, back_E_eta_er) * pow(back_engine_mass / back_mtom, back_E_K_er));
  }
  // Secondary structure mass
  double back_mass_secondary_structure = 0.0443 * back_mtom * config->back_chiozzotto_technology_factor.value();

  // Mass of covers
  const auto [back_C_cov, back_mtom_cov, back_w_to_s_cov, back_ar_cov, back_cos_lambda_cov, back_t_to_c_kink_cov, back_v_eas_design_cov,
              back_one_plus_taper_cov] = lookup_table_covers[back_material];
  double back_mass_covers = (back_k_e * back_C_cov * pow(back_mtom, back_mtom_cov) * pow(back_wing_loading, back_w_to_s_cov) * pow(back_aspect_ratio, back_ar_cov) *
                        pow(cos(back_quarter_chord_sweep), back_cos_lambda_cov) * pow(back_t_to_c_kink, back_t_to_c_kink_cov) *
                        pow(v_eas_design, back_v_eas_design_cov) * pow(1 + back_taper_ratio, back_one_plus_taper_cov)) *
                       config->back_chiozzotto_technology_factor.value();

  // Mass of spars and ribs
  const auto [back_C_sr, back_mtom_sr, back_w_to_s_sr, back_ar_sr, back_cos_lambda_sr, back_t_to_c_kink_sr, back_v_eas_design_sr, back_one_plus_taper_sr] =
      lookup_table_spars_and_ribs[back_material];
  double back_mass_spars_and_ribs = (back_C_sr * pow(back_mtom, back_mtom_sr) * pow(back_wing_loading, back_w_to_s_sr) * pow(back_aspect_ratio, back_ar_sr) *
                                pow(cos(back_quarter_chord_sweep), back_cos_lambda_sr) * pow(back_t_to_c_kink, back_t_to_c_kink_sr) *
                                pow(v_eas_design, back_v_eas_design_sr) * pow(1 + back_taper_ratio, back_one_plus_taper_sr)) *
                               config->back_chiozzotto_technology_factor.value();

  double back_wing_mass = back_mass_covers + back_mass_spars_and_ribs + back_mass_secondary_structure;

  myRuntimeInfo->out << "Calculated back wing mass ... " << back_wing_mass << "[kg]" << std::endl;
  myRuntimeInfo->out << "Back Wing mass spars and ribs ..." << back_mass_spars_and_ribs << "[kg]" << std::endl;
  myRuntimeInfo->out << "Back Wing mass covers  ..." << back_mass_covers << "[kg]" << std::endl;
  myRuntimeInfo->out << "Back Wing mass secondary structure ..." << back_mass_secondary_structure << "[kg]" << std::endl;

  
  myRuntimeInfo->out << "Overall Tandem Wing Set mass estimation by adding both wings masses calculated with Chiozotto method ..." << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;

  const double tandem_wing_set_mass = wing_mass + back_wing_mass;

  myRuntimeInfo->out << "Calculated overall tandem wing set mass ... " << tandem_wing_set_mass << "[kg]" << std::endl;

  data->wing_mass_properties.data["mass"] = tandem_wing_set_mass;
  //data->back_wing_mass_properties.data["mass"] = back_wing_mass;  //TODO kayra: olmamasi lazim
  
  
  // Calculate wing center of gravity of front wing according to Howe 2005 page 252
  myRuntimeInfo->out << "Calculate Center of Gravity of front wing ..." << std::endl;
  const double half_span_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  const double mean_chord = geom2::measure::reference_area(data->wing) / geom2::measure::span(data->wing);
  const double mac = geom2::measure::chord(data->wing, half_span_mac_position);
  const geom2::Point_3 leading_edge_point = geom2::measure::offset_LE(data->wing, half_span_mac_position); //TODO kayra: if twist this should be adjusted
  const double CoG_x = data->wing_position.x.value() + leading_edge_point.x() + 0.25 * mac +
                                                 0.1 * mean_chord;
  // Transform values into global coordinate system z-> -y z
  const double CoG_y = data->wing_position.y.value() + fabs(leading_edge_point.z()) -
                                                 fabs(leading_edge_point.z());
  const double CoG_z = data->wing_position.z.value() + leading_edge_point.y();

  myRuntimeInfo->out << "CG Position of front wing..." << std::endl;
  myRuntimeInfo->out << "x ... " << CoG_x << std::endl;
  myRuntimeInfo->out << "y ... " << CoG_y << std::endl;
  myRuntimeInfo->out << "z ... " << CoG_z << std::endl;




  // Calculate wing center of gravity of back wing according to Howe 2005 page 252
  myRuntimeInfo->out << "Calculate Center of Gravity of back wing..." << std::endl;
  const double back_half_span_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->back_wing);
  const double back_mean_chord = geom2::measure::reference_area(data->back_wing) / geom2::measure::span(data->back_wing);
  const double back_mac = geom2::measure::chord(data->back_wing, back_half_span_mac_position);
  const geom2::Point_3 back_leading_edge_point = geom2::measure::offset_LE(data->back_wing, back_half_span_mac_position); //TODO kayra: if twist this should be adjusted
  const double back_CoG_x = data->wing_position.x.value() + data->back_wing.origin.x() + back_leading_edge_point.x() + 0.25 * back_mac +
                                                 0.1 * back_mean_chord;
  // Transform values into global coordinate system z-> -y z
  const double back_CoG_y = data->wing_position.y.value() + data->back_wing.origin.y() + fabs(back_leading_edge_point.z()) -
                                                 fabs(back_leading_edge_point.z());
  const double back_CoG_z = data->wing_position.z.value() + data->back_wing.origin.z() + back_leading_edge_point.y();

  myRuntimeInfo->out << "CG Position of back wing..." << std::endl;
  myRuntimeInfo->out << "x ... " << back_CoG_x << std::endl;
  myRuntimeInfo->out << "y ... " << back_CoG_y << std::endl;
  myRuntimeInfo->out << "z ... " << back_CoG_z << std::endl;

 
  // Calculate center of gravity of tandem wing set using mass weighted CoG formula //TODO kayra
  myRuntimeInfo->out << "Calculate Center of Gravity of overall tandem wing set ..." << std::endl;
  const double wingset_CoG_x = (CoG_x * wing_mass + back_CoG_x * back_wing_mass) / tandem_wing_set_mass;
  const double wingset_CoG_y = (CoG_y * wing_mass + back_CoG_y * back_wing_mass) / tandem_wing_set_mass;
  const double wingset_CoG_z = (CoG_z * wing_mass + back_CoG_z * back_wing_mass) / tandem_wing_set_mass;

  
  data->wing_mass_properties.data["x"].set_value(wingset_CoG_x);
  data->wing_mass_properties.data["y"].set_value(wingset_CoG_y);
  data->wing_mass_properties.data["z"].set_value(wingset_CoG_z);

  myRuntimeInfo->out << "CG Position of Tandem Wing Set ..." << std::endl;
  myRuntimeInfo->out << "x ... " << data->wing_mass_properties.data["x"].value() << std::endl;
  myRuntimeInfo->out << "y ... " << data->wing_mass_properties.data["y"].value() << std::endl;
  myRuntimeInfo->out << "z ... " << data->wing_mass_properties.data["z"].value() << std::endl;

    
}


bool Wing::chiozzotto_wer_check_limits(std::tuple<double, double>& min_max, double value_to_check) {
  const auto [min, max] = min_max;
  if ((value_to_check > min && value_to_check < max) ||
      (fabs(value_to_check - min) < ACCURACY_MEDIUM || fabs(value_to_check - max) < ACCURACY_MEDIUM)) {
    return true;
  } else {
    return false;
  }
}

double Wing::get_control_device_area_in_fraction_of_wing_area() {
  double overall_control_device_area = 0.0;
  double wing_area = geom2::measure::reference_area(data->wing);
  for (auto& item : data->control_devices) {
    auto cd = geom2::transform::to_absolute(item, data->wing);
    overall_control_device_area += geom2::measure::area(cd);
  }
  return overall_control_device_area / wing_area;
}

double Wing::get_back_control_device_area_in_fraction_of_back_wing_area() {
  double overall_control_device_area = 0.0;
  double wing_area = geom2::measure::reference_area(data->back_wing);
  for (auto& item : data->back_control_devices) {
    auto cd = geom2::transform::to_absolute(item, data->back_wing);
    overall_control_device_area += geom2::measure::area(cd);
  }
  return overall_control_device_area / wing_area;
}

geom2::AirfoilSection Wing::get_rescaled_airfoil_geometry(std::tuple<double, double, std::string> wing_profile_and_thickness) {
  const auto [eta, t_to_c, wing_profile] = wing_profile_and_thickness;
  geom2::Polygon_2 shape = airfoils_library->get_airfoil(wing_profile);
  airfoils_library->copy_available_airfoil(wing_profile,rtIO->getAirfoilDataDir());
  geom2::AirfoilSection section(shape);
  double max_thickness_to_chord = geom2::measure::thickness_max(section) / section.get_chord_length();
  section.scale_thickness(t_to_c / max_thickness_to_chord);

  return section;
}

 geom2::Mesh Wing::combine_meshes(geom2::Mesh mesh1, const geom2::Mesh mesh2) {

    // Define the Surface_mesh as a new type for a 3D point mesh 
    typedef CGAL::Surface_mesh<CGAL::Simple_cartesian<double>::Point_3> Surface_mesh;

    // Map to store the correspondence between vertices from mesh2 to mesh1
    std::map<Surface_mesh::Vertex_index, Surface_mesh::Vertex_index> vmap;

    // Iterate over vertices of mesh2 and add them to mesh1
    for (const auto& v2 : mesh2.vertices()) {
        // Add vertex to mesh1 (this returns a reference to the new vertex)
        vmap[v2] = mesh1.add_vertex(mesh2.point(v2));
    }

    // Iterate over faces of mesh2 and add them to mesh1
    for (const auto& f2 : mesh2.faces()) {
        // Collect the vertices of the face in mesh2
        std::vector<Surface_mesh::Vertex_index> vertices;
        Surface_mesh::Halfedge_index h = mesh2.halfedge(f2);
        for (auto h_around = mesh2.halfedges_around_face(h).begin(); h_around != mesh2.halfedges_around_face(h).end(); ++h_around) {
            Surface_mesh::Vertex_index v2 = mesh2.source(*h_around);
            vertices.push_back(vmap[v2]);
        }
        // Add face to mesh1
        mesh1.add_face(vertices);
    }

    return mesh1;
}

}  // namespace cantilever
}  // namespace tandem

// Helper methods

