/**
 * \file cantileverWingDesignPlot.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Plots for cantileverWingDesign
 * \version 0.1
 * \date 2024-05-22
 *
 * @copyright Copyright (c) 2024
 *
 */

#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <matplot/matplot.h>
#include <moduleBasics/svg.h>

#include <cstdlib>
#include <map>
#include <deque>
#include <utility>
#include <array>

#include "cantileverTandemWingDesign.h"


/**
 * \brief Generate polygon from spar or controldevice
 *
 * \param surface aerodynamic surface object
 * \param object spar or control device object
 * \return std::tuple<std::vector<double>,std::vector<double>,std::string> coordinates and object name
 */
std::tuple<std::vector<double>,std::vector<double>,std::string> gen_polygon_tandem(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface, const geom2::MultisectionSurface<geom2::PolygonSection>& object);

/**
 * \brief RGB and alpha values to Matplot++ format
 *
 * \param r red value between 0 - 255
 * \param b blue value between 0 - 255
 * \param g green alue between 0 - 255
 * \param a alpha value between 0 (invisible) - 1 (solid)
 * \return std::array<float,4UL> matplot++ conform color format
 */
std::array<float,4UL> rgba_tandem(float r, float b, float g, float a) {
  return {1-a,r/255,b/255,g/255};
}

//TODO kayra: error : altough there is no kink. Calculated with kink
/* */
std::array<float,4UL> spar_color_tandem = rgba_tandem(99, 110, 114,1.0);

std::map<std::string,std::array<float,4UL>> control_device_colormap_tandem = {
  {"aileron",rgba_tandem(108, 92, 231,1.0)},
  {"slat",rgba_tandem(232, 67, 147,1.0)},
  {"fowler",rgba_tandem(0, 184, 148,1.0)},
  {"spoiler",rgba_tandem(253, 203, 110,1.0)}
};

std::tuple<std::vector<double>,std::vector<double>,std::string> gen_polygon_tandem(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface, const geom2::MultisectionSurface<geom2::PolygonSection>& object) {
  double half_span = 0.5 * geom2::measure::span(surface);
  /* Spar coordinates positive values normally*/
  double from = object.sections.at(0).origin.z() * half_span;
  double to = object.sections.at(1).origin.z() * half_span;

  /* Chordwise coordinates */
  double chord_from_fwd = object.sections.at(0).get_contour().vertex(0).x();
  double chord_from_aft = object.sections.at(0).get_contour().vertex(1).x();
  double chord_to_fwd = object.sections.at(1).get_contour().vertex(0).x();
  double chord_to_aft = object.sections.at(1).get_contour().vertex(1).x();

  auto f_chord_fwd = [chord_from_fwd,chord_to_fwd,half_span,from,to](double span_position) { return std::lerp(chord_from_fwd,chord_to_fwd,fabs(span_position-fabs(from))/(fabs(to)-fabs(from))); };
  auto f_chord_aft = [chord_from_aft,chord_to_aft,half_span,from,to](double span_position) { return std::lerp(chord_from_aft,chord_to_aft,fabs(span_position-fabs(from))/(fabs(to)-fabs(from))); };
  /* Add object from */
  std::vector<double> spanwise_coordinates{-fabs(from)};

  /* Add intermediate sections */
  for (auto& section : surface.sections) {
    if (fabs(section.origin.z()) > fabs(from) && fabs(section.origin.z()) < fabs(to) ) {
      spanwise_coordinates.push_back(section.origin.z());
    }
  }
  /* Add object to */
  spanwise_coordinates.push_back(-fabs(to));

  /**/
  std::vector<double> x;
  std::vector<double> y;
  /* From in to out*/
  for(auto it = spanwise_coordinates.begin(); it != spanwise_coordinates.end(); ++it) {
    y.push_back(fabs(*it));
    x.push_back(geom2::measure::offset_LE(surface,*it).x() + f_chord_fwd(*it)*geom2::measure::chord(surface,*it));
  }
  /* From out to in*/
  for(auto rit = spanwise_coordinates.rbegin(); rit != spanwise_coordinates.rend(); ++rit) {
    y.push_back(fabs(*rit));
    x.push_back(geom2::measure::offset_LE(surface,*rit).x() + f_chord_aft(*rit)*geom2::measure::chord(surface,*rit));
  }

  /* Close loop */
  y.push_back(y.front());
  x.push_back(x.front());




  return {x,y,object.name};
}
namespace tandem {
namespace cantilever {



fs::path Wing::plot_wing_planform() {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }

  auto coordinates_xy = geom2::transform::outline_2d(data->wing, geom2::Direction_3(0, 0, 1));
  std::deque<double> spanwise;
  std::deque<double> chordwise;
  auto half_span = 0.5 * geom2::measure::span(data->wing);
  auto sections = data->wing.sections;
  auto fuselage_half = geom2::measure::width(data->fuselage , data->wing_position.x.value())*0.5;

  /* Loop coordinates from out to inside for LE and TE */
  for (auto it = sections.rbegin(); it != sections.rend(); ++it) {
    spanwise.push_front(-it->origin.z());
    spanwise.push_back(-it->origin.z());
    chordwise.push_front(it->origin.x());
    chordwise.push_back(it->origin.x()+it->get_chord_length());
  }

  /* Get mac coordinates */
  auto mac_position_y = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  auto mac_position = geom2::measure::offset_LE(data->wing,mac_position_y);
  auto mac = geom2::measure::chord(data->wing,mac_position_y);



  /* Generate plot */
  std::vector<std::string> legend_strings;
  auto f = matplot::figure(true);
  f->title("Wing planform");
  auto ax = f->add_axes(false);
  ax->plot(std::vector<double>(spanwise.begin(),spanwise.end()), std::vector<double>(chordwise.begin(),chordwise.end()))->color("black").line_width(1.5);
  legend_strings.push_back("wing contour");
  matplot::hold(true);
  ax->plot({-mac_position.z(),-mac_position.z()},{mac_position.x(),mac_position.x()+mac})->color({0,128,128,128}).line_style(":").line_width(1.5);
  legend_strings.push_back("MAC");

  /* Plot spars */
  for(auto& spar : data->spars) {
    const auto [x,y,_] = gen_polygon_tandem(data->wing, spar);
    (void)_;
    ax->plot(y,x)->color(spar_color_tandem).line_width(1).line_style("--");
    legend_strings.push_back("spar");
  }
  for(auto& device : data->control_devices) {
    const auto [x,y,name] = gen_polygon_tandem(data->wing, device);
    ax->plot(y,x)->color(control_device_colormap_tandem[name]).line_width(1.5);
    legend_strings.push_back(device.name);
  }
  auto limits = ax->ylim();
  ax->plot(std::vector<double>({fuselage_half,fuselage_half}),std::vector<double>({limits[0]-2,limits[1]+2}))->color({0, 1, 0,0}).line_style("--").line_width(1.0);
  legend_strings.push_back("fuselage width");
  matplot::legend_handle lh;
  ax->xlabel("y [m]");
  ax->ylabel("x [m]");
  ax->limits_mode(matplot::manual);
  ax->axis(matplot::equal);
  ax->xlim({0,*std::max_element(spanwise.begin(), spanwise.end())*1.1});
  ax->legend(legend_strings);
  ax->legend()->location(matplot::legend::general_alignment::bottomright);
  ax->legend()->inside(true);
  ax->legend()->box(true);
  ax->legend()->font_size(8);


  fs::path plot_path = rtIO->getPlotDir() + "/wing_planfrom.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  system("pkill gnuplot");
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

fs::path Wing::plot_back_wing_planform() {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }

  auto coordinates_xy = geom2::transform::outline_2d(data->back_wing, geom2::Direction_3(0, 0, 1));
  std::deque<double> spanwise;
  std::deque<double> chordwise;
  auto half_span = 0.5 * geom2::measure::span(data->back_wing);
  auto sections = data->back_wing.sections;
  auto fuselage_half = geom2::measure::width(data->fuselage , (data->wing_position.x.value()+data->back_wing.origin.x()))*0.5;

  /* Loop coordinates from out to inside for LE and TE */
  for (auto it = sections.rbegin(); it != sections.rend(); ++it) {
    spanwise.push_front(-it->origin.z());
    spanwise.push_back(-it->origin.z());
    chordwise.push_front(it->origin.x());
    chordwise.push_back(it->origin.x()+it->get_chord_length());
  }

  /* Get mac coordinates */
  auto mac_position_y = geom2::measure::mean_aerodynamic_chord_position(data->back_wing);
  auto mac_position = geom2::measure::offset_LE(data->back_wing,mac_position_y);
  auto mac = geom2::measure::chord(data->back_wing,mac_position_y);



  /* Generate plot */
  std::vector<std::string> legend_strings;
  auto f = matplot::figure(true);
  f->title("Wing planform");
  auto ax = f->add_axes(false);
  ax->plot(std::vector<double>(spanwise.begin(),spanwise.end()), std::vector<double>(chordwise.begin(),chordwise.end()))->color("black").line_width(1.5);
  legend_strings.push_back("wing contour");
  matplot::hold(true);
  ax->plot({-mac_position.z(),-mac_position.z()},{mac_position.x(),mac_position.x()+mac})->color({0,128,128,128}).line_style(":").line_width(1.5);
  legend_strings.push_back("MAC");

  /* Plot spars */
  for(auto& spar : data->back_spars) {
    const auto [x,y,_] = gen_polygon_tandem(data->back_wing, spar);
    (void)_;
    ax->plot(y,x)->color(spar_color_tandem).line_width(1).line_style("--");
    legend_strings.push_back("spar");
  }
  for(auto& device : data->back_control_devices) {
    const auto [x,y,name] = gen_polygon_tandem(data->back_wing, device);
    ax->plot(y,x)->color(control_device_colormap_tandem[name]).line_width(1.5);
    legend_strings.push_back(device.name);
  }
  auto limits = ax->ylim();
  ax->plot(std::vector<double>({fuselage_half,fuselage_half}),std::vector<double>({limits[0]-2,limits[1]+2}))->color({0, 1, 0,0}).line_style("--").line_width(1.0);
  legend_strings.push_back("fuselage width");
  matplot::legend_handle lh;
  ax->xlabel("y [m]");
  ax->ylabel("x [m]");
  ax->limits_mode(matplot::manual);
  ax->axis(matplot::equal);
  ax->xlim({0,*std::max_element(spanwise.begin(), spanwise.end())*1.1});
  ax->legend(legend_strings);
  ax->legend()->location(matplot::legend::general_alignment::bottomright);
  ax->legend()->inside(true);
  ax->legend()->box(true);
  ax->legend()->font_size(8);


  fs::path plot_path = rtIO->getPlotDir() + "/back_wing_planfrom.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  system("pkill gnuplot");
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

fs::path Wing::plot_thickness_and_twist_distribution() {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }
  std::vector<double> eta;
  std::vector<double> thickness_to_chord_ratio;
  std::vector<double> twist;
  double half_span = geom2::measure::span(data->wing) * 0.5;
  for (auto section : data->wing.sections) {
    eta.push_back(-section.origin.z() / half_span);
    thickness_to_chord_ratio.push_back(section.get_thickness_scale() * geom2::measure::thickness_max(section) /
                                       section.get_chord_length());
    twist.push_back(section.get_twist_angle());
  }

  auto f = matplot::figure(true);

  auto ax1 = f->add_subplot(2, 1, 0);
  ax1->plot(eta, thickness_to_chord_ratio, "g-")->color("black").line_width(1.5);
  // ax1->xlabel("\u03B7");
  ax1->x_axis().visible(true);
  ax1->grid(true);
  ax1->ylabel("t/c");

  auto ax2 = f->add_subplot(2, 1, 1);
  ax2->plot(eta, twist, "b-")->color("black").line_style("--").line_width(1.5);
  ax2->ylabel("\u03F5 [\u00b0]");
  ax2->xlabel("\u03B7");
  ax2->grid(true);
  /*
  // Plot the first set of data
  ax1->plot(eta, thickness_to_chord_ratio, "g-")->use_y2(false).color("black").line_width(1.5);
  ax1->xlabel("\u03B7");
  ax1->ylabel("t/c");
  ax1->limits_mode_automatic();
  ax1->legend({"t/c"});

  matplot::hold(true);
  // Create a secondary y-axis
  auto ax2 = f->add_axes(ax1, false, false);
  ax2->plot(eta, twist, "b-")->use_y2(true).color("green").line_style("--").line_width(1.5);
  ax2->y2_axis().limits({-1, 1});
  // ax2->visible(false);
  ax2->grid(true);
  ax2->y_axis().color("black");
  ax2->y2_axis().label("\u03F5 [\u00b0]");
  ax2->legend({"t/c", "\u03F5"});

  // Set the secondary y-axis invisible except for the y-axis itself
  ax2->x_axis().visible(true);
  ax2->y2_axis().color("black");
  // Adjust limits and ticks if necessary
  ax2->ylim({0, 0.2});
  ax2->y2lim({-1, 1});
  */
  fs::path plot_path = rtIO->getPlotDir() + "/front_wing_thickness_and_twist_distribution.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  system("pkill gnuplot");
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

fs::path Wing::plot_back_thickness_and_twist_distribution() {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }
  std::vector<double> eta;
  std::vector<double> thickness_to_chord_ratio;
  std::vector<double> twist;
  double half_span = geom2::measure::span(data->back_wing) * 0.5;
  for (auto section : data->back_wing.sections) {
    eta.push_back(-section.origin.z() / half_span);
    thickness_to_chord_ratio.push_back(section.get_thickness_scale() * geom2::measure::thickness_max(section) /
                                       section.get_chord_length());
    twist.push_back(section.get_twist_angle());
  }

  auto f = matplot::figure(true);

  auto ax1 = f->add_subplot(2, 1, 0);
  ax1->plot(eta, thickness_to_chord_ratio, "g-")->color("black").line_width(1.5);
  // ax1->xlabel("\u03B7");
  ax1->x_axis().visible(true);
  ax1->grid(true);
  ax1->ylabel("t/c");

  auto ax2 = f->add_subplot(2, 1, 1);
  ax2->plot(eta, twist, "b-")->color("black").line_style("--").line_width(1.5);
  ax2->ylabel("\u03F5 [\u00b0]");
  ax2->xlabel("\u03B7");
  ax2->grid(true);
  /*
  // Plot the first set of data
  ax1->plot(eta, thickness_to_chord_ratio, "g-")->use_y2(false).color("black").line_width(1.5);
  ax1->xlabel("\u03B7");
  ax1->ylabel("t/c");
  ax1->limits_mode_automatic();
  ax1->legend({"t/c"});

  matplot::hold(true);
  // Create a secondary y-axis
  auto ax2 = f->add_axes(ax1, false, false);
  ax2->plot(eta, twist, "b-")->use_y2(true).color("green").line_style("--").line_width(1.5);
  ax2->y2_axis().limits({-1, 1});
  // ax2->visible(false);
  ax2->grid(true);
  ax2->y_axis().color("black");
  ax2->y2_axis().label("\u03F5 [\u00b0]");
  ax2->legend({"t/c", "\u03F5"});

  // Set the secondary y-axis invisible except for the y-axis itself
  ax2->x_axis().visible(true);
  ax2->y2_axis().color("black");
  // Adjust limits and ticks if necessary
  ax2->ylim({0, 0.2});
  ax2->y2lim({-1, 1});
  */
  fs::path plot_path = rtIO->getPlotDir() + "/back_wing_thickness_and_twist_distribution.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  system("pkill gnuplot");
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}
/**
 * \brief Plot airfoils of front wing
 *
 * \return fs::path
 */
std::vector<fs::path> Wing::plot_airfoils() {
  /* Plot active */
  if (!rtIO->plotOn) {
    return {""};
  }
  std::map<std::string, std::vector<geom2::Point_2>> airfoils_data{};
  std::vector<fs::path> airfoil_plots_path;

  for (auto section : data->wing.sections) {
    if (airfoils_data.find(section.name) != airfoils_data.end()) {
      continue;
    } else {
      airfoils_data.insert({section.name, section.get_contour().vertices()});
    }
  }
  /* loop over airfoils */

  for (const auto& airfoil_data : airfoils_data) {
    std::vector<double> x{}, y{};
    for (auto coord : airfoil_data.second) {
      x.push_back(coord.x());
      y.push_back(coord.y());
    }

    {
      auto f = matplot::figure(true);
      auto h = f->current_axes()->plot(x, y);
      h->line_width(1.5).color("black");

      auto ax = f->current_axes();
      ax->title(std::format("Airfoil - {}", airfoil_data.first));
      matplot::xlabel("\u03B7");
      ax->ylabel("z/c");
      ax->xlabel("\u03B7");
      ax->grid(true);
      ax->ylim({-0.2, 0.2});
      ax->xlim({0, 1});
      ax->axis(matplot::equal);

      fs::path plot_path = rtIO->getPlotDir() + "/" + airfoil_data.first + "_airfoil.svg";
      matplot::save(f, plot_path.string(), "svg");
      airfoil_plots_path.push_back(fs::relative(plot_path, fs::path(rtIO->getHtmlDir())));
      std::this_thread::sleep_for(std::chrono::milliseconds(500));
      system("pkill gnuplot");
    }
  }
  return airfoil_plots_path;
}

/**
 * \brief Plot airfoils of back wing
 *
 * \return fs::path
 */
std::vector<fs::path> Wing::plot_back_airfoils() {
  /* Plot active */
  if (!rtIO->plotOn) {
    return {""};
  }
  std::map<std::string, std::vector<geom2::Point_2>> airfoils_data{};
  std::vector<fs::path> airfoil_plots_path;

  for (auto section : data->back_wing.sections) {
    if (airfoils_data.find(section.name) != airfoils_data.end()) {
      continue;
    } else {
      airfoils_data.insert({section.name, section.get_contour().vertices()});
    }
  }
  /* loop over airfoils */

  for (const auto& airfoil_data : airfoils_data) {
    std::vector<double> x{}, y{};
    for (auto coord : airfoil_data.second) {
      x.push_back(coord.x());
      y.push_back(coord.y());
    }

    {
      auto f = matplot::figure(true);
      auto h = f->current_axes()->plot(x, y);
      h->line_width(1.5).color("black");

      auto ax = f->current_axes();
      ax->title(std::format("Airfoil - {}", airfoil_data.first));
      matplot::xlabel("\u03B7");
      ax->ylabel("z/c");
      ax->xlabel("\u03B7");
      ax->grid(true);
      ax->ylim({-0.2, 0.2});
      ax->xlim({0, 1});
      ax->axis(matplot::equal);

      fs::path plot_path = rtIO->getPlotDir() + "/" + airfoil_data.first + "_airfoil.svg";
      matplot::save(f, plot_path.string(), "svg");
      airfoil_plots_path.push_back(fs::relative(plot_path, fs::path(rtIO->getHtmlDir())));
      std::this_thread::sleep_for(std::chrono::milliseconds(500));
      system("pkill gnuplot");
    }
  }
  return airfoil_plots_path;
}


}  // namespace cantilever
}  // namespace tandem
